from collections import OrderedDict
from pc23 import xrange
from .exceptions import OrmRequestError, OrmAttributeTypeMistake, OrmModelClassMismatch, OrmDefinitionError, \
    OrmParameterError, OrmModelDeclarationError
from .models.meta import OrmModelMeta
from .models.base import OrmModelAbstract, OrmModelAliased, CompositeOrmModel
from .attributes.base import DataAttributeAbstract, AttributeAliased, OrmComparator, ColumnAttributeAbstract, \
    OrmComparatorMixin
from .attributes.columns import ColumnIntegerAbstract, ColumnBooleanAbstract, ColumnDatetimeAbstract
from .attributes.relationships import ForeignAttr
from .funcs import DbFunctionAbstract, RequestWhere, RequestHaving, RequestLimit, RequestOrderBy, \
    RequestGroupBy, RequestNamedColumn, RequestTableBelongedColumn, RequestJoinAbstract, RawSql, left_join_, \
    right_join_, inner_join_, full_join_, t_, or_
from .results import OrmResult
from .manage import manager
from .utils import split_search_phrase

REQUEST_SELECT = 'select'
REQUEST_UPDATE = 'update'
REQUEST_INSERT = 'insert'
REQUEST_REPLACE = 'replace'
REQUEST_DELETE = 'delete'
REQUEST_UNDELETE = 'undelete'


class OrmRequest(object):
    def __init__(self, request_type, *args, **kwargs):
        self._type = request_type.lower()
        self._aliases = set()
        self._aliases_tbl = set()
        self._aliases_col = set()
        self._attrs = dict()
        self._sources = list()
        self._columns = list()
        self._joins = list()
        self._models = OrderedDict()
        self._tbl_aliases = dict()
        self._col_aliases = dict()
        self._where = None
        self._having = None
        self._limit = None
        self._order_by = None
        self._group_by = None
        self._union_order_by = None
        self._union_group_by = None
        self._union_limit = None
        self._is_finished = False
        self._union = None
        self._req_columns = None
        self._req_srcs = list()
        self._src_belongs_to = dict()
        self._c_belongs_to = dict()
        self._requesting_sources = list()
        self._inserted_id = None
        self._resulting_model = None
        self._foreign_attrs = dict()
        self._autojoins = dict()
        self._related_objects = OrderedDict()
        self._related_obj_attrs = list()
        self.session = kwargs.pop('session', None)
        self.params = dict()
        self.ignore = kwargs.pop('ignore', False)
        self.distinct = kwargs.pop('distinct', False)
        self.high_priority = kwargs.pop('high_priority', False)
        self.straight_join = kwargs.pop('straight_join', False)
        self.sql_big_result = kwargs.pop('sql_big_result', False)
        self.sql_small_result = kwargs.pop('sql_small_result', False)
        self.sql_buffer_result = kwargs.pop('sql_buffer_result', False)
        self.sql_calc_found_rows = kwargs.pop('sql_calc_found_rows', False)
        self.sql_cache = kwargs.pop('sql_cache', False)
        self.sql_no_cache = kwargs.pop('sql_no_cache', False) if not self.sql_cache else False
        self.for_update = kwargs.pop('for_update', False)
        self.lock_in_share_mode = kwargs.pop('lock_in_share_mode', False)
        self.deleted = kwargs.pop('with_deleted', False)
        self.values = kwargs.pop('values', None)
        self.hard_delete = bool(kwargs.pop('hard_delete', False))
        self.with_values = kwargs.pop('with_values', None)
        self.with_default = kwargs.pop('with_default', None)
        self.textual_options = bool(kwargs.pop('textual_options', False))
        self.only_own = bool(kwargs.pop('only_own', False))
        if self.values and self._type not in ('update', 'insert', 'replace'):
            raise OrmRequestError("'values' might be set for UPDATE, INSERT or REPLACE requests only!")
        if self._type in ('update', 'insert') and not self.values:
            raise OrmRequestError("'values' must be set for UPDATE, INSERT or REPLACE request!")
        if self.ignore and self._type != 'insert':
            raise OrmRequestError("'ignore' can only be set for INSERT command!")

        for arg in args:
            if isinstance(arg, (OrmModelMeta, OrmModelAbstract, OrmModelAliased)):
                self._add_source(arg)
                self._requesting_sources.append(arg)
            elif isinstance(arg, RequestJoinAbstract):
                self._add_join(arg)
            elif isinstance(arg, (DataAttributeAbstract, DbFunctionAbstract)):
                self._resulting_model = CompositeOrmModel
                self._add_attribute(arg)
            elif isinstance(arg, (AttributeAliased, RequestTableBelongedColumn)):
                self._resulting_model = CompositeOrmModel
                self._add_attribute(arg)
            elif isinstance(arg, OrmComparator):
                self._add_where(arg)
            elif isinstance(arg, RequestWhere):
                self._add_where(arg)
            elif isinstance(arg, RequestHaving):
                self._add_having(arg)
            elif isinstance(arg, RequestLimit):
                self._set_limit(arg)
            elif isinstance(arg, RequestOrderBy):
                self._add_orderby(arg)
            elif isinstance(arg, RequestGroupBy):
                self._add_groupby(arg)
            else:
                raise OrmRequestError(
                    "Unsupported type of argument given for the request of type '%s'!" % str(type(arg))
                )

    def _add_source(self, src):
        if src in self._sources and not isinstance(src, OrmModelAliased):
            raise OrmRequestError(
                "OrmModel must be named using .as_() method if this model should be used several times"
                " it the query!"
            )
        elif src in self._sources:
            raise OrmRequestError(
                "The same OrmModel, with same naming, cannot be specified in the request more than once!"
            )
        if isinstance(src, OrmModelAliased):
            if src.alias in self._aliases:
                raise OrmRequestError("Alias is already defined: '%s'!" % str(src.alias))
            self._register_table_alias(src.alias, src)
        self._sources.append(src)
        self._req_srcs.append(src)

    def _add_join(self, j, belongs_to=None):
        src = j.source
        if src in self._sources and not isinstance(src, OrmModelAliased):
            raise OrmRequestError(
                "OrmModel must be named using .as_() method if this model should be used several times"
                " it the query!"
            )
        elif src in self._sources:
            raise OrmRequestError(
                "The same OrmModel, with same naming, cannot be specified in the request more than once!"
            )
        if isinstance(src, OrmModelAliased):
            if src.alias in self._aliases:
                raise OrmRequestError("Alias is already defined: '%s'!" % str(src.alias))
            self._register_table_alias(src.alias, src)
        self._joins.append(j)
        req_src = belongs_to if belongs_to else j
        if req_src not in self._req_srcs:
            self._req_srcs.append(req_src)

    def _add_attribute(self, c):
        if isinstance(c, ColumnAttributeAbstract) and self._type == 'select':
            k = c.tbl_k or c.model_k
            c = c.make_read(self)
            c.tbl_k = k
            c.model_k = k
        elif isinstance(c, RequestTableBelongedColumn):
            c.make_read(self)
        if isinstance(c, AttributeAliased):
            if c.alias in self._aliases:
                raise OrmRequestError("Alias is already defined: '%s'!" % str(c.alias))
            self._register_column_alias(c.alias, c.c)
        else:
            self._attrs[c] = None
        self._columns.append(c)

    def _add_where(self, *expr):
        for arg in expr:
            if isinstance(arg, RequestWhere):
                arg = arg.args
            if self._where is None:
                self._where = RequestWhere(arg)
            elif isinstance(self._where, RequestWhere):
                self._where.extend_with(arg)
            else:
                raise OrmRequestError(
                    "Unexpected type of value for filter 'WHERE' of the request: %s" % str(type(self._where))
                )

    def _add_having(self, *expr):
        for arg in expr:
            if isinstance(arg, RequestHaving):
                arg = arg.args
            if self._having is None:
                self._having = RequestHaving(arg)
            elif isinstance(self._having, RequestHaving):
                self._having.extend_with(arg)
            else:
                raise OrmRequestError(
                    "Unexpected type of value for filter 'HAVING' of the request: %s" % str(type(self._having))
                )

    def _set_limit(self, arg):
        if self._union is None or self._union is True:
            self._limit = arg
        else:
            self._union_limit = arg

    def _add_orderby(self, arg):
        if self._union is None or self._union is True:
            if self._order_by is None:
                self._order_by = list()
            self._order_by.append(arg)
        else:
            if self._union_order_by is None:
                self._union_order_by = list()
            self._union_order_by.append(arg)

    def _add_groupby(self, arg):
        if self._union is None or self._union is True:
            if self._group_by is None:
                self._group_by = list()
            self._group_by.append(arg)
        else:
            if self._union_group_by is None:
                self._union_group_by = list()
            self._union_group_by.append(arg)

    def _fillup_columns(self):
        if self._col_aliases or self._columns:
            return
        for src in self._models:
            if self._requesting_sources \
                    and not (src in self._requesting_sources
                             or (isinstance(src, OrmModelAliased) and src.alias in self._requesting_sources)):
                continue
            model = src.model if isinstance(src, OrmModelAliased) else src
            attrs = model.__meta__.get_attributes_for_select() \
                if self._type == 'select' \
                else model.__meta__.get_attributes_for_write()
            for c in attrs:
                if issubclass(self.resulting_model, CompositeOrmModel) \
                        and not getattr(c, 'in_composite_results', True):
                    continue
                alias = c.tbl_k or c.model_k
                c_ = RequestTableBelongedColumn(alias, src.alias) if isinstance(src, OrmModelAliased) else c
                self._add_attribute(c_)

    def _register_table_alias(self, alias, model):
        self._aliases.add(alias)
        self._aliases_tbl.add(alias)
        self._tbl_aliases[alias] = model
        self._models[model] = alias

    def _register_column_alias(self, alias, c):
        self._aliases.add(alias)
        self._aliases_col.add(alias)
        self._attrs[c] = alias
        self._col_aliases[alias] = c

    def _get_free_table_alias(self, prefix):
        i = len(self._aliases_tbl)
        n = "%s%i" % (str(prefix), i)
        while n in self._aliases:
            i += 1
            n = "%s%i" % (str(prefix), i)
        return n

    def _get_free_column_alias(self, prefix, c=None):
        if c is not None and isinstance(c, ColumnAttributeAbstract):
            c_src = c.model
            tbl_name = self._models[c_src]
            col_name = c.tbl_k
            col_alias = "%s_%s" % (tbl_name, col_name)
            if col_alias not in self._aliases:
                return col_alias
        elif c is not None and isinstance(c, RequestTableBelongedColumn):
            tbl_name = c.tbl_name
            model = self._tbl_aliases[tbl_name]
            if isinstance(model, OrmModelAliased):
                model = model.model
            k = c.k
            c = getattr(model, k)
            col_name = c.tbl_k
            if col_name:
                col_alias = "%s_%s" % (tbl_name, col_name)
                if col_alias not in self._aliases:
                    return col_alias
        i = len(self._aliases_col) + 1
        n = "%s%i" % (str(prefix), i)
        while n in self._aliases:
            i += 1
            n = "%s%i" % (str(prefix), i)
        return n

    def _name_unnamed_sources(self):
        for src in self._sources:
            if isinstance(src, OrmModelAliased):
                continue
            tbl_alias = self._get_free_table_alias('t')
            self._register_table_alias(tbl_alias, src)
        for j in self._joins:
            if isinstance(j.source, OrmModelAliased):
                continue
            src = j.source
            tbl_alias = self._get_free_table_alias('t')
            self._register_table_alias(tbl_alias, src)

    def _name_unnamed_columns(self):
        for c in self._columns:
            if isinstance(c, AttributeAliased):
                continue
            col_alias = self._get_free_column_alias('c', c)
            self._register_column_alias(col_alias, c)

    def get_valueparam_by_column(self, c):
        if c in self.values:
            value = self.values[c]
        elif isinstance(c, DataAttributeAbstract):
            k = c.model_k
            if k in self.values:
                value = self.values[k]
            else:
                return None
        else:
            return None
        pn = c.model_k if isinstance(c, DataAttributeAbstract) else str(c)
        if not pn:
            return None
        self.params[pn] = value
        return pn

    def _sql_from(self):
        if self._type in ('insert', 'replace') and len(self._sources) != 1:
            raise OrmRequestError("INSERT requests must have ONE OrmModel to be specified!")
        p = list()
        for src in self._sources:
            tbl_alias = src.alias if isinstance(src, OrmModelAliased) else self._models[src]
            tbl_name = src.model.table if isinstance(src, OrmModelAliased) else src.table
            p.append("%s AS %s" % (self.session.quote_name(tbl_name), tbl_alias)
                     if self._type not in ('insert', 'replace', 'update', 'delete')
                     else "%s" % self.session.quote_name(tbl_name))
        return ", ".join(p)

    def _sql_joins(self):
        p = list()
        for j in self._joins:
            src = j.source
            tbl_alias = src.alias if isinstance(src, OrmModelAliased) else self._models[src]
            tbl_name = src.model.table if isinstance(src, OrmModelAliased) else src.table
            join_sql = j.join_type
            j_cond = j.condition
            sql_cond = self.resolve_expr(j_cond)
            f_softdelete = self._source_softdelete_filter(src)
            if f_softdelete:
                sql_softdelete = self.resolve_expr(f_softdelete, allow_alias=False)
                sql_cond = sql_softdelete + " AND " + sql_cond
            p.append("%s %s AS %s ON %s" % (join_sql, self.session.quote_name(tbl_name), tbl_alias, sql_cond))
        return " ".join(p)

    def _sql_sources(self):
        gsrc = list()
        gsrc.append(self._sql_from())
        if self._type in ('insert', 'replace') and self._joins:
            raise OrmRequestError("Cannot use JOINs in the INSERT or REPLACE request!")
        joins = self._sql_joins()
        if joins:
            gsrc.append(joins)
        return " ".join(gsrc)

    def _sql_insert_columns(self):
        self._req_columns = list()
        p = list()
        for c in self._columns:
            q = self.sql_column(c, permit_c=True, permit_f=False)
            qp = self._sql_write_param(c)
            if not qp or not q:
                continue
            p.append(q)
            self._req_columns.append(c)
        if not p:
            return None
        return ", ".join(p)

    def _sql_insert_values(self):
        p = list()
        for c in self._columns:
            q = self.sql_column(c, permit_c=True, permit_f=False)
            if q:
                qp = self._sql_write_param(c)
                if not qp:
                    continue
                p.append(qp)
        if not p:
            return None
        return ", ".join(p)

    def _sql_write_param(self, c):
        c_ = c.make_write(self) if isinstance(c, ColumnAttributeAbstract) else c
        if isinstance(c_, DbFunctionAbstract):
            return c_.resolve(self)
        elif isinstance(c_, ColumnAttributeAbstract):
            return c_.make_write(self)
        return c_

    def _sql_columns(self):
        self._req_columns = list()
        p = list()
        for c in self._columns:
            q = self.sql_column(c, permit_c=True, permit_f=True)
            if q:
                if self._type in ('update', 'insert', 'replace'):
                    qp = self._sql_write_param(c)
                    if not qp:
                        continue
                    q += "=" + qp
                p.append(q)
                self._req_columns.append(c)
        if not p:
            return None
        return ", ".join(p)

    def _sql_where(self):
        if self._where is None:
            return ""
        q_ = self.resolve_expr(self._where.args, allow_alias=False)
        if not q_:
            return ""
        return q_

    def _sql_having(self):
        if self._having is None:
            return ""
        q_ = self.resolve_expr(self._having.args, allow_alias=True)
        if not q_:
            return ""
        return q_

    def _sql_limit(self, value=None):
        value = value or self._limit
        if not value:
            return ""
        if not isinstance(value, RequestLimit):
            raise OrmRequestError("Request._limit must be a type of (RequestLimit)!")
        return "%i OFFSET %i" % (value.limit, value.offset) \
            if value.offset \
            else str(value.limit)

    def _sql_orderby(self):
        if not self._order_by:
            return ""
        p = list()
        for arg in self._order_by:
            if not isinstance(arg, RequestOrderBy):
                raise OrmRequestError("Request._order_by element must be a type of (RequestOrderBy)!")
            c = arg.c
            if isinstance(c, ColumnAttributeAbstract):
                c = c.make_order_by(self)
            p.append("%s %s" % (self._resolve_attr(c, allow_alias=True), arg.d))
        return ", ".join(p)

    def _sql_groupby(self):
        if not self._group_by:
            return ""
        p = list()
        for arg in self._group_by:
            if not isinstance(arg, RequestGroupBy):
                raise OrmRequestError("Request._group_by element must be a type of (RequestGroupBy)!")
            c = arg.c
            if isinstance(c, ColumnAttributeAbstract):
                c = c.make_group_by(self)
            p.append(self._resolve_attr(c, allow_alias=True))
        return ", ".join(p)

    def _sql_union_limit(self, value=None):
        value = value or self._union_limit
        if not value:
            return ""
        if not isinstance(value, RequestLimit):
            raise OrmRequestError("Request._limit must be a type of (RequestLimit)!")
        return "%i OFFSET %i" % (value.limit, value.offset) \
            if value.offset \
            else str(value.limit)

    def _sql_union_orderby(self):
        if not self._union_order_by:
            return ""
        p = list()
        for arg in self._union_order_by:
            if not isinstance(arg, RequestOrderBy):
                raise OrmRequestError("Request._order_by element must be a type of (RequestOrderBy)!")
            c = arg.c
            if isinstance(c, ColumnAttributeAbstract):
                c = c.make_order_by(self)
            p.append("%s %s" % (self._resolve_attr(c, allow_alias=True), arg.d))
        return ", ".join(p)

    def _sql_union_groupby(self):
        if not self._union_group_by:
            return ""
        p = list()
        for arg in self._union_group_by:
            if not isinstance(arg, RequestGroupBy):
                raise OrmRequestError("Request._group_by element must be a type of (RequestGroupBy)!")
            c = arg.c
            if isinstance(c, ColumnAttributeAbstract):
                c = c.make_group_by(self)
            p.append(self._resolve_attr(c, allow_alias=True))
        return ", ".join(p)

    def _resolve_attr(self, item, allow_alias=False, table_name=None):
        if item is None:
            return "NULL"
        elif isinstance(item, (ColumnAttributeAbstract, RequestTableBelongedColumn, DbFunctionAbstract)) \
                and item in self._attrs \
                and allow_alias:
            return str(self._attrs[item])
        elif isinstance(item, AttributeAliased) and allow_alias:
            return item.alias
        elif isinstance(item, ColumnAttributeAbstract):
            c_src = item.model
            tbl_name = self._models[c_src] if not table_name else table_name
            col_name = item.tbl_k
            return "%s.%s" % (self.session.quote_name(tbl_name), self.session.quote_name(col_name)) \
                if self._type not in ('update', 'delete') \
                else self.session.quote_name(col_name)
        elif isinstance(item, RequestTableBelongedColumn):
            k = item.k
            tbl_name = item.tbl_name
            if tbl_name not in self._tbl_aliases:
                raise OrmRequestError(
                    "Table with alias name '%s' is not defined in the request, for column '%s'!" % (tbl_name, k)
                )
            c_model = self._tbl_aliases[tbl_name]
            c_model_ = c_model.model if isinstance(c_model, OrmModelAliased) else c_model
            if not hasattr(c_model_, k):
                raise OrmRequestError(
                    "OrmModel '%s', which is associated with table alias '%s' in the request, has no"
                    " attribute '%s' declared in it!" % (c_model_.__name__, tbl_name, k)
                )
            c_ = getattr(c_model_, k)
            if isinstance(c_, ColumnAttributeAbstract):
                col_name = c_.tbl_k
                return "%s.%s" % (self.session.quote_name(tbl_name), self.session.quote_name(col_name))
            elif isinstance(c_, DbFunctionAbstract):
                return c_.resolve(self, table_name=tbl_name)
            else:
                raise OrmRequestError(
                    "Attribute '%s' of OrmModel '%s', declared in the request with table alias '%s', is"
                    " not a type of (Column|DbFunction) and cannot be queried!" % (k, c_model_.__name__, tbl_name)
                )
        elif isinstance(item, OrmRequest):
            query_ = item.sql
            params_ = item.params
            for k in params_:
                v = params_[k]
                k_ = 'p%i' % len(self.params)
                kf_ = "%(" + k + ")s"
                kt_ = "%(" + k_ + ")s"
                query_ = query_.replace(kf_, kt_)
                self.params[k_] = v
            return "(" + query_ + ")"
        elif isinstance(item, DbFunctionAbstract):
            return item.resolve(self, table_name=table_name)
        elif isinstance(item, RawSql):
            return str(item.value)
        elif isinstance(item, RequestNamedColumn):
            return item.column_name
        pn = "p%i" % len(self.params)
        self.params[pn] = item
        sn = "%(" + pn + ")s"
        return sn

    def resolve_expr(self, expr, allow_alias=False, table_name=None):
        def _true(expr_):
            if isinstance(expr_, ColumnAttributeAbstract):
                return expr_.true_
            return expr_

        def _false(expr_):
            if isinstance(expr_, ColumnAttributeAbstract):
                return expr_.false_
            return "NOT %s" % str(expr_)

        def _is_null(expr_):
            return "%s IS NULL" % expr_

        def _is_not_null(expr_):
            return "%s IS NOT NULL" % expr_

        def _in(expr_, values_, isnot_=False):
            rv_ = list()
            if isinstance(values_, OrmRequest):
                query_ = expr_.sql
                params_ = expr_.params
                for k in params_:
                    v = params_[k]
                    k_ = 'p%i' % len(self.params)
                    kf_ = "%(" + k + ")s"
                    kt_ = "%(" + k_ + ")s"
                    query_ = query_.replace(kf_, kt_)
                    self.params[k_] = v
                rv = query_
            elif isinstance(values_, (list, tuple)):
                if not isinstance(values_, (list, tuple)):
                    values_ = [values_, ]
                for v_ in values_:
                    vq_ = str(v_)
                    pn_ = "p%i" % (len(self.params))
                    rv_.append("%(" + pn_ + ")s")
                    self.params[pn_] = vq_
                rv = ", ".join(rv_)
            elif isinstance(values_, str):
                rv = values_
            else:
                raise OrmRequestError("'IN' and 'NOT IN' must have a (list|tuple) or OrmRequest as parameter!")
            q_not_ = " NOT" if isnot_ else ""
            return str(expr_) + q_not_ + " IN (" + rv + ")"

        def _not_in(expr_, values_):
            return _in(expr_, values_, True)

        def _cmp(expr_, cmp_, value_):
            if value_ is None:
                if cmp_ == '=':
                    return _is_null(expr_)
                elif cmp_ == '!=':
                    return _is_not_null(expr_)
                else:
                    raise OrmRequestError("SQL comparition with NULL may be only 'IS NULL' or 'IS NOT NULL'!")
            elif isinstance(value_, (list, tuple)):
                if cmp_ == '=':
                    return _in(expr_, value_)
                elif cmp_ == '!=':
                    return _not_in(expr_, value_)
                else:
                    raise OrmRequestError("SQL comparition with array may be only 'IN' or 'NOT IN'!")
            return expr_ + " " + cmp_ + " " + value_

        if expr is True or expr is False:
            raise OrmRequestError(
                "Incorrect filter is set. Probably trying to compare non database reflected attributes"
                " with some value. Only database reflected columns (columns, directly pointing to the"
                " corresponding database table columns) can be used in the query filters in ORM."
            )
        if not expr:
            return ""
        if not isinstance(expr, (list, tuple)):
            expr = ['AND', expr, ]
        if isinstance(expr[0], str) and expr[0] not in ('EXISTS', '!EXISTS', 'AND', 'OR'):
            raise OrmRequestError('Incorrect joins condition for filter in query: %s' % expr[0])
        elif not isinstance(expr[0], str):
            expr = ['AND'] + list(expr)
        if expr[0] == 'EXISTS':
            if len(expr) != 2:
                raise OrmRequestError("Incorrect filter with EXISTS modifier!")
            r_ = self.resolve_expr(expr[1], table_name=table_name)
            if not r_:
                return ""
            return "EXISTS(" + r_ + ")"
        elif expr[0] == '!EXISTS':
            if len(expr) != 2:
                raise OrmRequestError("Incorrect filter with NOT EXISTS modifier!")
            r_ = self.resolve_expr(expr[1], table_name=table_name)
            if not r_:
                return ""
            return "NOT EXISTS(" + r_ + ")"
        if expr[0] not in ('AND', 'OR'):
            expr = ['AND'] + list(expr)
        fj_ = " %s " % expr[0]
        fc_ = expr[1:]
        r_ = list()
        for f_ in fc_:
            if not isinstance(f_, (OrmComparator, tuple, list)):
                f_ = OrmComparator(f_, 'TRUE') if not isinstance(f_, OrmComparatorMixin) else f_.true_
            if isinstance(f_, OrmComparator):
                f_ = f_.resolve(self.session)
            if isinstance(f_, OrmComparator):
                item_ = f_.item
                if isinstance(item_, ColumnAttributeAbstract):
                    item_ = item_.make_filter(self)
                item_ = self._resolve_attr(item_, allow_alias=allow_alias, table_name=table_name)
                comparator_ = f_.cmp
                if comparator_ == 'ISNULL':
                    r_.append(_is_null(item_))
                    continue
                elif comparator_ == '!ISNULL':
                    r_.append(_is_not_null(item_))
                    continue
                elif comparator_ == 'TRUE':
                    r_.append(_true(item_))
                    continue
                elif comparator_ == 'FALSE':
                    r_.append(_false(item_))
                    continue
                right_ = f_.value
                if not isinstance(right_, (list, tuple)):
                    right_ = self._resolve_attr(right_, table_name=table_name)
                if comparator_ == 'IN':
                    r_.append(_in(item_, right_))
                    continue
                elif comparator_ == '!IN':
                    r_.append(_not_in(item_, right_))
                    continue
                elif isinstance(right_, (list, tuple)):
                    raise OrmRequestError("Cannot use (list|tuple) for comparators other than IN or NOT IN!")
                else:
                    r_.append(_cmp(item_, comparator_, right_))
            else:
                rf_ = self.resolve_expr(f_, table_name=table_name)
                if rf_:
                    rf_ = ("(" + rf_ + ")") if len(f_) > 1 else rf_
                    r_.append(rf_)
        if not r_:
            return ""
        return str(fj_).join(r_) if len(r_) > 1 else str(r_[0])

    def ensure_column(self, column):
        if isinstance(column, RequestNamedColumn):
            if column.column_name in self._col_aliases:
                raise OrmRequestError("Column is not defined in the request: '%s'!" % column.column_name)
        if column not in self._columns:
            self._add_attribute(column)

    def _handle_sql_column(self, c, *args):
        sql_ = list()
        for arg in args:
            sql_.append(self.session.quote_name(arg))
        sql = ".".join(sql_)
        if not self.textual_options:
            return sql
        options = getattr(c, 'options', None)
        if not options:
            return sql
        if not isinstance(options, dict):
            raise OrmDefinitionError("attribute's 'options' must be a (dict) type!")
        sql_case = list()
        for k in options:
            v = options[k]
            pn_k = "p%i" % len(self.params)
            self.params[pn_k] = k
            pn_v = "p%i" % len(self.params)
            self.params[pn_v] = v
            sql_case.append("WHEN %(" + pn_k + ")s THEN %(" + pn_v + ")s")
        sql = "(CASE " + str(sql) + " " + " ".join(sql_case) + " ELSE " + str(sql) + " END)"
        return sql

    def sql_column(self, c, permit_c=True, permit_f=True, alias=True, table_name=None):
        if self._type in ('insert', 'replace', 'update', 'delete'):
            if not isinstance(c, ColumnAttributeAbstract):
                return None
            return "%s" % self.session.quote_name(c.tbl_k)
        tbl_name = table_name
        if (isinstance(c, ForeignAttr) or (isinstance(c, AttributeAliased) and isinstance(c.c, ForeignAttr))) \
                and self._type == 'select':
            if isinstance(c, AttributeAliased):
                c = c.c
            c.validate()
            rel_obj_k = c.related_k
            rel_model_name = c.model.__name__
            rel_aj_name = "%s_%s" % (rel_model_name, rel_obj_k)
            if rel_aj_name not in self._autojoins:
                return None
            alias_name = self._attrs[c] if c in self._attrs else None
            rel_c = getattr(c.model, rel_obj_k)
            rel_c.validate()
            foreign_model = rel_c.foreign_model
            ref_k = c.referenced_k
            ref_c = getattr(foreign_model, ref_k)
            if not isinstance(ref_c, ColumnAttributeAbstract):
                return None
            col_name = ref_c.tbl_k
            tbl_name = self._autojoins[rel_aj_name]
            return "%s AS %s" % (self._handle_sql_column(rel_c, tbl_name, col_name), alias_name) \
                if alias and alias_name \
                else self._handle_sql_column(rel_c, tbl_name, col_name)
        if isinstance(c, DataAttributeAbstract) and not isinstance(c, ColumnAttributeAbstract):
            return None
        if isinstance(c, RequestNamedColumn):
            return str(c.column_name)
        if isinstance(c, ColumnAttributeAbstract) and not permit_c:
            return None
        if isinstance(c, DataAttributeAbstract):
            c_src = c.model
            tbl_name = self._models[c_src] if not table_name else table_name
            col_name = c.tbl_k
            alias_name = self._attrs[c] if c in self._attrs else None
            return "%s AS %s" % (self._handle_sql_column(c, tbl_name, col_name), alias_name) \
                if alias and alias_name and self._type == 'select' \
                else self._handle_sql_column(c, tbl_name, col_name)
        if isinstance(c, AttributeAliased):
            c_ = c.c
            alias_name = c.alias
        else:
            c_ = c
            alias_name = self._attrs[c] if c in self._attrs else None
        if isinstance(c_, ColumnAttributeAbstract) and permit_c:
            c_src = c_.model
            tbl_name = self._models[c_src] if not table_name else table_name
            col_name = c_.tbl_k
            return "%s AS %s" % (self._handle_sql_column(c_, tbl_name, col_name), alias_name) \
                if alias and alias_name and self._type == 'select' \
                else self._handle_sql_column(c_, tbl_name, col_name)
        if isinstance(c_, RequestTableBelongedColumn):
            k = c_.k
            tbl_name = c_.tbl_name if not table_name else table_name
            if tbl_name not in self._tbl_aliases:
                raise OrmRequestError(
                    "Table with alias name '%s' is not defined in the request, for column '%s'!" % (tbl_name, k)
                )
            c_model = self._tbl_aliases[tbl_name]
            c_model_ = c_model.model if isinstance(c_model, OrmModelAliased) else c_model
            if not hasattr(c_model_, k):
                raise OrmRequestError(
                    "OrmModel '%s', which is associated with table alias '%s' in the request, has no"
                    " attribute '%s' declared in it!" % (c_model_.__name__, tbl_name, k)
                )
            c_ = getattr(c_model_, k) if c_.c is None else c_.c
            if isinstance(c_, ColumnAttributeAbstract) and permit_c:
                col_name = c_.tbl_k
                return "%s AS %s" % (self._handle_sql_column(c_, tbl_name, col_name), alias_name) \
                    if alias and alias_name and self._type == 'select' \
                    else self._handle_sql_column(tbl_name, col_name)
        if isinstance(c_, DbFunctionAbstract) and permit_f:
            expr_sql = c_.resolve(self, table_name=tbl_name)
            return "%s AS %s" % (str(expr_sql), alias_name) \
                if alias and alias_name and self._type == 'select' \
                else str(expr_sql)
        return None

    @staticmethod
    def _source_softdelete_filter(src):
        model = src.model if isinstance(src, OrmModelAliased) else src
        c_softdelete = model.__meta__.soft_delete
        if not c_softdelete:
            return None
        fc_ = getattr(t_(src), c_softdelete.model_k) if isinstance(src, OrmModelAliased) else c_softdelete
        if isinstance(c_softdelete, ColumnBooleanAbstract):
            return fc_.false_
        elif isinstance(c_softdelete, ColumnDatetimeAbstract):
            return fc_.false_
        elif isinstance(c_softdelete, ColumnIntegerAbstract):
            return fc_.notnull_
        else:
            raise OrmAttributeTypeMistake(
                "Soft-delete key might be of type (Boolean|Date|DateTime|Timestamp|Integer) only!"
            )

    def _sql_softdelete_set(self):
        src = self._sources[0]
        model = src.model if isinstance(src, OrmModelAliased) else src
        c_softdelete = model.__meta__.soft_delete
        if not c_softdelete:
            return None
        c_ = getattr(t_(src), c_softdelete.model_k) if isinstance(src, OrmModelAliased) else c_softdelete
        q_ = self.sql_column(c_, permit_c=True, permit_f=False, alias=False)
        if isinstance(c_softdelete, ColumnBooleanAbstract):
            return q_ + "=TRUE"
        elif isinstance(c_softdelete, ColumnDatetimeAbstract):
            return q_ + "=NOW()"
        elif isinstance(c_softdelete, ColumnIntegerAbstract):
            return q_ + "=NULL"
        else:
            raise OrmAttributeTypeMistake(
                "Soft-delete key might be of type (Boolean|Date|DateTime|Timestamp|Integer) only!"
            )

    def _handle_softdelete_select(self):
        if self.deleted or self._type == 'delete':
            return
        if self._type in ('insert', 'replace'):
            return
        for src in self._sources:
            f_ = self._source_softdelete_filter(src)
            if f_ is None:
                continue
            self._add_where(f_)

    def _handle_foreign_attrs(self):
        if not self._columns:
            return
        for c in self._columns:
            if isinstance(c, AttributeAliased):
                c = c.c
            if not isinstance(c, ForeignAttr):
                continue
            c.validate()
            rel_obj_k = c.related_k
            rel_model_name = c.model.__name__
            rel_aj_name = "%s_%s" % (rel_model_name, rel_obj_k)
            if rel_aj_name not in self._autojoins:
                j_alias = self._get_free_table_alias('t')
                self._autojoins[rel_aj_name] = j_alias
                rel_c = getattr(c.model, rel_obj_k)
                rel_c.validate()
                j_src = rel_c.foreign_model.as_(j_alias)
                j_on_args = list()
                foreign_model = rel_c.foreign_model
                foreign_c = rel_c.foreign_c
                for x in xrange(len(foreign_c.columns)):
                    if foreign_c.model == c.model:
                        fc_ = getattr(foreign_model, foreign_c.ref_columns[x])
                        tfc_ = getattr(t_(j_alias, foreign_model), fc_.tbl_k)
                        lc_ = getattr(c.model, foreign_c.columns[x])
                    else:
                        fc_ = getattr(foreign_model, foreign_c.columns[x])
                        tfc_ = getattr(t_(j_alias, foreign_model), fc_.tbl_k)
                        lc_ = getattr(c.model, foreign_c.ref_columns[x])
                    j_on_args.append(tfc_ == lc_)
                self.left_join(j_src, *j_on_args)

    def _handle_related_attrs(self):
        if self._type != 'select':
            return
        resulting_model = self.resulting_model
        if issubclass(resulting_model, CompositeOrmModel):
            return
        if self.only_own:
            return
        attrs = resulting_model.__meta__.get_related_to_one_attributes()
        for c in attrs:
            if c.always_deferred:
                continue
            c.validate()
            j_alias = self._get_free_table_alias('t')
            j_src = c.foreign_model.as_(j_alias)
            j_on_args = list()
            foreign_model = c.foreign_model
            foreign_c = c.foreign_c
            self._related_objects[c.model_k] = (j_alias, foreign_model, list(), foreign_c)
            for x in xrange(len(foreign_c.columns)):
                if foreign_c.model == c.model:
                    fc_ = getattr(foreign_model, foreign_c.ref_columns[x])
                    tfc_ = getattr(t_(j_alias, foreign_model), fc_.tbl_k)
                    lc_ = getattr(c.model, foreign_c.columns[x])
                else:
                    fc_ = getattr(foreign_model, foreign_c.columns[x])
                    tfc_ = getattr(t_(j_alias, foreign_model), fc_.tbl_k)
                    lc_ = getattr(c.model, foreign_c.ref_columns[x])
                j_on_args.append(tfc_ == lc_)
            self.left_join(j_src, *j_on_args)

            attrs = foreign_model.__meta__.get_attributes_for_select()
            for fc in attrs:
                fc_alias = fc.tbl_k or fc.model_k
                fc_ = RequestTableBelongedColumn(fc_alias, j_alias, fc)
                self._add_attribute(fc_)
                self._related_obj_attrs.append(fc_)
                self._related_objects[c.model_k][2].append(fc_)

    def _finish(self, for_union=False):
        if self._is_finished:
            raise OrmRequestError(
                "Trying to modify request which is already finished and SQL query already prepared!"
            )
        self._is_finished = True
        if for_union:
            self._union = True
        self._name_unnamed_sources()
        self._handle_foreign_attrs()
        self._fillup_columns()
        self._handle_related_attrs()
        self._name_unnamed_columns()
        self._handle_softdelete_select()

    def _sql_delete(self):
        return self._sql_harddelete() \
            if self._sources[0].__meta__.soft_delete is None or self.hard_delete \
            else self._sql_softdelete()

    def _sql_softdelete(self):
        q_sources = self._sql_sources()
        q_where = self._sql_where()
        q_set = self._sql_softdelete_set()
        p = list()
        p.append("UPDATE")
        p.append(q_sources)
        p.append("SET")
        p.append(q_set)
        if q_where:
            p.append("WHERE")
            p.append(q_where)
        return " ".join(p)

    def _sql_harddelete(self):
        q_sources = self._sql_sources()
        q_where = self._sql_where()
        p = list()
        p.append("DELETE")
        # ps_ = list()
        # for src in self._sources:
        #     tbl_alias = src.alias if isinstance(src, OrmModelAliased) else self._models[src]
        #     ps_.append(tbl_alias)
        # p.append(",".join(ps_))
        p.append("FROM")
        p.append(q_sources)
        if q_where:
            p.append("WHERE")
            p.append(q_where)
        return " ".join(p)

    def _sql_write(self, cmd):
        q_sources = self._sql_sources()
        q_where = self._sql_where() if cmd not in ('insert', 'replace') else False
        q_columns = self._sql_columns() if cmd != 'insert' else self._sql_insert_columns()
        p = list()
        p.append(cmd.upper())
        if self.ignore:
            p.append('IGNORE')
        if cmd in ('insert', 'replace'):
            p.append('INTO')
        p.append(q_sources)
        if cmd == 'insert':
            p.append("(" + q_columns + ")")
            p.append("VALUES")
            p.append("(" + self._sql_insert_values() + ")")
        else:
            p.append("SET")
            p.append(q_columns)
            if q_where:
                p.append("WHERE")
                p.append(q_where)
        if not q_columns:
            return None
        return " ".join(p)

    def _sql_select(self, for_union=False, page=None, per_page=None, limit=None):
        if page is not None and not per_page:
            raise OrmRequestError(
                "When selecting only a page from an entire list - the parameter 'per_page' is required"
                " along with 'page'!"
            )
        if page and limit:
            raise OrmRequestError(
                "Parameters 'page' and 'limit' are mutual exclusive and cannot be specified both!"
            )
        q_sources = self._sql_sources()
        q_columns = self._sql_columns()
        q_where = self._sql_where()
        q_having = self._sql_having()
        if page is not None or limit is not None:
            if not self._union or self._union is True:
                q_limit = "" \
                    if limit is False \
                    else (self._sql_limit(limit)
                          if limit and isinstance(limit, RequestLimit)
                          else (self._sql_limit(RequestLimit(per_page * (page - 1), per_page)) if page else ""))
            else:
                q_limit = ""
        else:
            q_limit = self._sql_limit()
        q_orderby = self._sql_orderby()
        q_groupby = self._sql_groupby()
        p = list()
        p.append("SELECT")
        if self.distinct:
            p.append("DISTINCT")
        if self.high_priority and not for_union:
            p.append("HIGH_PRIORITY")
        if self.straight_join:
            p.append("STRAIGHT_JOIN")
        if self.sql_small_result:
            p.append("SQL_SMALL_RESULT")
        if self.sql_big_result:
            p.append("SQL_BIG_RESULT")
        if self.sql_buffer_result and not for_union:
            p.append("SQL_BUFFER_RESULT")
        if self.sql_cache and not for_union:
            p.append("SQL_CACHE")
        elif self.sql_no_cache and not for_union:
            p.append("SQL_NO_CACHE")
        if self.sql_calc_found_rows and not for_union:
            p.append("SQL_CALC_FOUND_ROWS")
        p.append(q_columns)
        p.append("FROM")
        p.append(q_sources)
        if q_where:
            p.append("WHERE")
            p.append(q_where)
        if q_groupby:
            p.append("GROUP BY")
            p.append(q_groupby)
        if q_having:
            p.append("HAVING")
            p.append(q_having)
        if q_orderby:
            p.append("ORDER BY")
            p.append(q_orderby)
        if q_limit:
            p.append("LIMIT")
            p.append(q_limit)
        if self.for_update and not for_union:
            p.append("FOR UPDATE")
        elif self.lock_in_share_mode and not for_union:
            p.append("LOCK IN SHARE MODE")
        if not q_columns:
            return None
        sql = " ".join(p)

        if self._union is not None and self._union is not True and isinstance(self._union, list):
            sql = "(" + sql + ")"
            p = list()
            p.append(sql)
            for union in self._union:
                p.append("UNION" if not union[1] else "UNION ALL")
                p.append("(" + union[0].union_request(self) + ")")
            q_union_groupby = self._sql_union_groupby()
            q_union_orderby = self._sql_union_orderby()
            if page is not None or limit is not None:
                q_union_limit = "" \
                    if limit is False \
                    else (self._sql_union_limit(limit)
                          if limit and isinstance(limit, RequestLimit)
                          else (self._sql_union_limit(RequestLimit(per_page * (page - 1), per_page)) if page else ""))
            else:
                q_union_limit = self._sql_union_limit()
            if q_union_groupby:
                p.append("GROUP BY")
                p.append(q_union_groupby)
            # if q_having:
            #     p.append("HAVING")
            #     p.append(q_having)
            if q_union_orderby:
                p.append("ORDER BY")
                p.append(q_union_orderby)
            if q_union_limit:
                p.append("LIMIT")
                p.append(q_union_limit)
            sql = " ".join(p)

        return sql

    def union_request(self, context):
        self.session = context.session
        if not self._is_finished:
            self._finish(for_union=True)
        union_sql = self._sql_select(for_union=True)
        for k in self.params:
            if k not in context.params:
                context.params[k] = self.params[k]
                continue
            pn = "pu%i_" % len(context.params)
            rpn_f = "%(" + k + ")s"
            rpn_t = "%(" + pn + ")s"
            context.params[pn] = self.params[k]
            union_sql = union_sql.replace(rpn_f, rpn_t)
        return union_sql

    def _sql(self, page=None, per_page=None, limit=None):
        if not self._is_finished:
            self._finish()
        if self._type == 'select':
            return self._sql_select(page=page, per_page=per_page, limit=limit)
        elif self._type in ('update', 'insert', 'replace'):
            return self._sql_write(self._type)
        elif self._type == 'delete':
            return self._sql_delete()
        return None

    def _require_session(self, session=None):
        from .session import OrmSessionAbstract
        session = session or self.session
        if not session:
            model_ = None
            for src in self._sources:
                ms_ = src.model.__meta__.session if isinstance(src, OrmModelAliased) else src.__meta__.session
                if ms_ is not None:
                    model_ = src.model if isinstance(src, OrmModelAliased) else src
                    break
            session = manager.get_session(session=session, model=model_)
        if not session:
            raise OrmRequestError("No database connection session is specified!")
        if not isinstance(session, OrmSessionAbstract):
            raise OrmRequestError("Database connection session ('session') must be a type of (OrmSessionAbstract)!")
        self.session = session
        return session

    @property
    def sql(self):
        return self._sql()

    def get(self, *args):
        """ This method offers an opportunity to set a set of sources (Models) and/or attributes to be loaded
        from the database. """
        if not args:
            raise OrmRequestError("'get' method of OrmRequest requires a set of sources and/or attributes!")
        sources = list()
        for arg in args:
            if isinstance(arg, (OrmModelMeta, OrmModelAbstract, OrmModelAliased)):
                sources.append(arg)
            elif isinstance(arg, (DataAttributeAbstract, DbFunctionAbstract)):
                self._resulting_model = CompositeOrmModel
                self._add_attribute(arg)
            elif isinstance(arg, (AttributeAliased, RequestTableBelongedColumn)):
                self._resulting_model = CompositeOrmModel
                self._add_attribute(arg)
        if sources:
            self._requesting_sources = sources
        return self

    @property
    def resulting_model(self):
        return self._resulting_model \
               or (self._req_srcs[0]
                   if (len(self._req_srcs) == 1 or len(self._requesting_sources) == 1)
                   and not self.textual_options and not self._union
                   else CompositeOrmModel)

    def go(self, session=None, page=None, per_page=None, limit=None):
        """ Executes the previously prepared request (this one). """
        session = self._require_session(session)
        self.session = session
        query = self._sql(page=page, per_page=per_page, limit=limit)
        params = self.params
        if self._type != 'select':
            session.execute(query, params)
            return
        rows = session.query(query, params).all() if query else []
        result = OrmResult()
        result.session = session
        result.request = self
        result.query = query
        result.params = params
        result.rows = rows
        result.page = page
        result.per_page = per_page
        columns2sources = dict()
        for c in self._columns:
            if isinstance(c, (ColumnAttributeAbstract, DbFunctionAbstract)):
                c_src = c.model
            elif isinstance(c, RequestTableBelongedColumn):
                c_src = self._tbl_aliases[c.tbl_name]
            else:
                c_src = None
            if c_src is not None and c_src in self._src_belongs_to:
                c_src = self._src_belongs_to[c_src]
            columns2sources[c] = c_src
        instance_cls = self.resulting_model
        sz_col = len(self._req_columns) - len(self._related_obj_attrs)
        for row in rows:
            kw = dict()
            for x in xrange(sz_col):
                c = self._req_columns[x]
                value = row[x]
                src = columns2sources[c]
                k = None
                if isinstance(c, AttributeAliased):
                    k = c.alias
                    c = c.c
                if isinstance(c, RequestTableBelongedColumn):
                    model = self._tbl_aliases[c.tbl_name]
                    model_ = model.model if isinstance(model, OrmModelAliased) else model
                    c_ = getattr(model_, c.k)
                else:
                    c_ = c
                k = k or getattr(c_, 'model_k', None)
                if not k:
                    k = self._attrs[c]
                elif instance_cls is CompositeOrmModel and src is not None:
                    if len(self._req_srcs) != 1 and len(self._requesting_sources) != 1:
                        model_name = src.alias if isinstance(src, OrmModelAliased) else src.__name__
                        k = "%s__%s" % (model_name, k)
                kw[k] = value, c_, src
            kw['__textual__'] = self.textual_options
            instance = self._create_loaded_instance(instance_cls, None, **kw)
            x = sz_col
            for rel_k in self._related_objects:
                finstance_cls = self._related_objects[rel_k][1]
                fattrs = self._related_objects[rel_k][2]
                relation_c = self._related_objects[rel_k][3]
                sz_fcol = len(fattrs)
                kw = dict()
                for n in xrange(sz_fcol):
                    c = self._req_columns[x]
                    value = row[x]
                    src = columns2sources[c]
                    k = c.k if isinstance(c, RequestTableBelongedColumn) else c.model_k
                    kw[k] = value, c, src
                    x += 1
                f_pk = finstance_cls.__meta__.get_primary_key()
                is_existing = False
                for cpk_ in f_pk:
                    kpk_ = cpk_.model_k
                    if kpk_ not in kw:
                        continue
                    vpk_ = kw[kpk_]
                    if vpk_[0] is not None:
                        is_existing = True
                        break
                if not is_existing:
                    continue
                kw['__textual__'] = self.textual_options
                related_instance = self._create_loaded_instance(finstance_cls, (instance, relation_c), **kw)
                rel_c = instance.__getattrabs__(rel_k)
                rel_c.set_value(instance, rel_k, related_instance)
            result.objects.append(instance)
        return result

    def _create_loaded_instance(self, __model__, __parent__=None, **kwargs):
        if isinstance(self.with_values, dict):
            for k in self.with_values:
                kwargs[k] = self.with_values[k]
        if isinstance(self.with_default, dict):
            for k in self.with_default:
                if k in kwargs:
                    continue
                kwargs[k] = self.with_default[k]
        instance = __model__()
        instance.__existing__ = True
        instance.__session__ = self.session
        instance.__parent__ = __parent__
        instance.__load__(**kwargs)
        return instance

    def count(self, session=None):
        if self._type != 'select':
            raise OrmRequestError("Method 'count()' can be called for selecting requests only!")
        session = self._require_session(session)
        q = "SELECT COUNT(1) FROM (" + self._sql(limit=False) + ") AS __tcount__"
        r = session.query(q, self.params).one()
        if not r:
            return None
        return int(r[0])

    def using(self, session):
        self.session = session

    def with_values(self, **kwargs):
        if self._type != 'select':
            raise OrmRequestError("'with_values' can be used with SELECT requests only!")
        if not isinstance(self.with_values, dict):
            self.with_values = dict()
        for k in kwargs:
            self.with_values[k] = kwargs[k]

    def with_default(self, **kwargs):
        if self._type != 'select':
            raise OrmRequestError("'with_default' can be used with SELECT requests only!")
        if not isinstance(self.with_default, dict):
            self.with_default = dict()
        for k in kwargs:
            self.with_default[k] = kwargs[k]

    def textual(self):
        self.textual_options = True
        return self

    def where(self, *expr):
        self._add_where(*expr)
        return self

    def having(self, *expr):
        self._add_having(*expr)
        return self

    def filter(self, *expr):
        return self.where(*expr)

    def join(self, model, *args):
        return self.left_join(model, *args)

    def left_join(self, model, *args):
        self._add_join(left_join_(model, *args))
        return self

    def right_join(self, model, *args):
        self._add_join(right_join_(model, *args))
        return self

    def inner_join(self, model, *args):
        self._add_join(inner_join_(model, *args))
        return self

    def full_join(self, model, *args):
        self._add_join(full_join_(model, *args))
        return self

    def union(self, *args):
        if self._type != 'select':
            raise OrmRequestError("cannot UNION when request is not SELECT!")
        for arg in args:
            if not isinstance(arg, OrmRequest):
                raise OrmRequestError("'union' method requires another OrmRequest to be given!")
            if self._union is None or self._union is True:
                self._union = list()
            self._union.append((arg, False))
        return self

    def union_all(self, *args):
        if self._type != 'select':
            raise OrmRequestError("cannot UNION ALL when request is not SELECT!")
        for arg in args:
            if not isinstance(arg, OrmRequest):
                raise OrmRequestError("'union_all' method requires another OrmRequest to be given!")
            if self._union is None or self._union is True:
                self._union = list()
            self._union.append((arg, True))
        return self

    def limit(self, *args):
        self._set_limit(RequestLimit(*args))
        return self

    def order_by(self, *args):
        if len(args) == 2 \
                and isinstance(args[1], str) \
                and args[1].upper() in ('ASC', 'DESC') \
                and isinstance(args[0], (ColumnAttributeAbstract, AttributeAliased, RequestTableBelongedColumn, str, int)):
            self._add_orderby(RequestOrderBy(args[0], args[1]))
            return self
        for arg in args:
            if isinstance(arg, (list, tuple)):
                if len(arg) != 2:
                    raise OrmRequestError(
                        "ORDER BY may be set using attributes or a (list|tuple) type array, first element of"
                        " which is an attribute and the second is a direction: ASC or DESC!"
                    )
                c = arg[0]
                d = arg[1]
                self._add_orderby(RequestOrderBy(c, d))
            else:
                self._add_orderby(RequestOrderBy(arg))
        return self

    def group_by(self, *args):
        if len(args) == 2 \
                and isinstance(args[1], str) \
                and args[1].upper() in 'ASC, DESC' \
                and isinstance(args[0], (ColumnAttributeAbstract, AttributeAliased, RequestTableBelongedColumn)):
            self._add_groupby(RequestGroupBy(args[0], args[1]))
            return self
        for arg in args:
            if isinstance(arg, (list, tuple)):
                if len(arg) != 2:
                    raise OrmRequestError(
                        "GROUP BY may be set using attributes or a (list|tuple) type array, first element of"
                        " which is an attribute and the second is a direction: ASC or DESC!"
                    )
                c = arg[0]
                d = arg[1]
                self._add_groupby(RequestGroupBy(c, d))
            else:
                self._add_groupby(RequestGroupBy(arg))
        return self

    @property
    def request_type(self):
        return self._type

    @property
    def first(self):
        if not self._limit:
            self.limit(1)
        return self.go().first

    @property
    def last(self):
        return self.go().last

    def one(self, x=None):
        if x is None:
            return self.first
        if not self._limit:
            self.limit(x, 1)
            return self.go().first
        return self.go().one(x)


class OrmSelectRequest(OrmRequest):
    def __init__(self, *args, **kwargs):
        super(OrmSelectRequest, self).__init__(REQUEST_SELECT, *args, **kwargs)


class OrmUpdateRequest(OrmRequest):
    def __init__(self, *args, **kwargs):
        super(OrmUpdateRequest, self).__init__(REQUEST_UPDATE, *args, **kwargs)


class OrmInsertRequest(OrmRequest):
    def __init__(self, *args, **kwargs):
        super(OrmInsertRequest, self).__init__(REQUEST_INSERT, *args, **kwargs)

    @property
    def inserted_id(self):
        if self._inserted_id:
            return self._inserted_id
        elif self.session.dbms_type == 'mysql':
            r = self.session.query("SELECT LAST_INSERT_ID()").one()
            if not r:
                raise Exception("Cannot retrieve LAST_INSERT_ID from DBMS")
            if not r[0]:
                raise Exception("Cannot retrieve LAST_INSERT_ID from DBMS, got %s" % str(r[0]))
            self._inserted_id = int(r[0])
            return int(r[0])
        elif self.session.dbms_type == 'postgresql':
            if self._sources[0].__meta__.auto_increment is None:
                raise OrmModelClassMismatch(
                    "OrmModel '%s' has no auto-increment key defined!" % self._sources[0].__class__.__name__
                )
            tbl_name = self._sources[0].table
            column_name = self._sources[0].__meta__.get_autoincrement().tbl_k
            q = "SELECT CURRVAL(pg_get_serial_sequence('%s', '%s'));" % (tbl_name, column_name)
            r = self.session.query(q).one()
            if not r:
                raise Exception("Cannot retrieve LAST_INSERT_ID from DBMS")
            if not r[0]:
                raise Exception("Cannot retrieve LAST_INSERT_ID from DBMS, got %s" % str(r[0]))
            self._inserted_id = int(r[0])
            return int(r[0])
        return None


class OrmReplaceRequest(OrmRequest):
    def __init__(self, *args, **kwargs):
        super(OrmReplaceRequest, self).__init__(REQUEST_REPLACE, *args, **kwargs)


class OrmDeleteRequest(OrmRequest):
    def __init__(self, *args, **kwargs):
        super(OrmDeleteRequest, self).__init__(REQUEST_DELETE, *args, **kwargs)


class OrmUndeleteRequest(OrmRequest):
    def __init__(self, *args, **kwargs):
        super(OrmUndeleteRequest, self).__init__(REQUEST_UNDELETE, *args, **kwargs)


class OrmSearchRequest(OrmRequest):
    def __init__(self, model, what, *args, **kwargs):
        super(OrmSearchRequest, self).__init__(REQUEST_SELECT, model, *args, **kwargs)
        if not isinstance(what, str):
            raise OrmParameterError("'find' method requires a search phrase of type of (str)!")
        findable_attrs = model.__meta__.get_findable_attributes()
        if not findable_attrs:
            raise OrmModelDeclarationError(
                "cannot 'find' over Model which there are no 'findable' attributes declared for it!"
            )
        phrases = split_search_phrase(what)
        w = list()
        for c in findable_attrs:
            if c.strict_find:
                w.append(c.in_(phrases))
            else:
                for phrase in phrases:
                    w.append(c.find_(phrase))
        self._add_where(or_(*w))





# -*- coding: utf8 -*-
""" Different functions (including MySQL functions mapping) and service classes. """

from .attributes.base import AttributeAbstract, OrmComparatorMixin, OrmComparator
from .exceptions import OrmParameterError, OrmDefinitionError


class UncastableColumn(object):
    """ A virtual class that used in functions when declaring _attrs which are castable itself, to
    prevent recursion when function resolved column which returns a function which resolves an
    argument which is castable column again and again and again. """
    def __init__(self, c):
        self.c = c


class RawSql(object):
    """ A class, used to inform the engine that this is a textual name of the requesting column
    or other textual raw query """
    def __init__(self, value):
        self.value = value


class RequestFilterAbstract(object):
    """ Abstract class used by condition filtering classes like RequestWhere, RequestHaving, etc. """
    def __init__(self, *args):
        if not args:
            raise OrmParameterError("Filter instance require at least one filter to be set!")
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = args[0]
        self.args = list(args)
        if self.args[0] not in ('AND', 'OR'):
            self.args = and_(*self.args)

    def extend_with(self, *args):
        if len(args) == 1 and isinstance(args[0], RequestFilterAbstract):
            args = args[0].args
        if self.args[0] == 'AND':
            self.args = list(self.args) + list(args)
        else:
            self.args = and_(self.args, args)


class RequestWhere(RequestFilterAbstract):
    """ SQL 'WHERE' conditions declaration """
    pass


class RequestHaving(RequestFilterAbstract):
    """ SQL 'HAVING' conditions declaration """
    pass


class RequestOrderBy(object):
    """ SQL 'ORDER BY' operator declaration """
    def __init__(self, c, d='ASC'):
        if isinstance(c, RequestOrderBy):
            self.c = c.c
            self.d = c.d
            return
        if not isinstance(d, str) or d.lower() not in ('asc', 'desc'):
            raise TypeError("Request's ORDER BY must be set using (<Column>, [ASC|DESC]) format!")
        self.c = c
        self.d = d.upper()


class RequestGroupBy(object):
    """ SQL 'GROUP BY' operator declaration """
    def __init__(self, c, d='ASC'):
        if isinstance(c, RequestGroupBy):
            self.c = c.c
            self.d = c.d
            return
        if not isinstance(d, str) or d.lower() not in ('asc', 'desc'):
            raise TypeError("Request's GROUP BY must be set using (<Column>, [ASC|DESC]) format!")
        self.c = c
        self.d = d.upper()


class RequestLimit(object):
    """ SQL 'LIMIT' operator declaration """
    def __init__(self, *args):
        if len(args) > 2:
            raise AttributeError("Request limit must be set with (limit) or (offset, limit) syntax!")
        if not args:
            raise AttributeError("Request limit must be set with (limit) or (offset, limit) syntax!")
        if len(args) == 1:
            if not isinstance(args[0], int):
                raise TypeError("Request limit must be set using integer value(s)!")
            self.offset = None
            self.limit = args[0]
        else:
            if not isinstance(args[0], int) or not isinstance(args[1], int):
                raise TypeError("Request limit must be set using integer value(s)!")
            self.offset = args[0]
            self.limit = args[1]


class RequestTableBelongedColumn(OrmComparatorMixin, object):
    def __init__(self, k, tbl_name, c=None):
        self.k = k
        self.tbl_name = tbl_name
        self.c = c

    def __hash__(self):
        return super(OrmComparatorMixin, self).__hash__()

    def as_(self, alias):
        from .attributes.base import AttributeAliased
        return AttributeAliased(self, alias)

    def __repr__(self):
        return "TableColumn(%s.%s)" % (self.tbl_name, self.k) \
            if self.c is None \
            else "TableColumn::%s(%s.%s)" % (self.c.__class__.__name__, self.tbl_name, self.k)

    def resolve_comparator_(self, cmp, session):
        if self.c is None:
            return super(RequestTableBelongedColumn, self).resolve_comparator_(cmp, session)
        return self.c.resolve_comparator_(cmp, session)

    def make_read(self, context):
        from .attributes.columns import ColumnAttributeAbstract
        if self.c is None or not isinstance(self.c, ColumnAttributeAbstract):
            return
        self.c = self.c.make_read(context)


class RequestAliasedSource(object):
    def __init__(self, tablename, model=None):
        self.tablename = tablename
        self.model = model

    def __getattribute__(self, key):
        if key.startswith('__') and key.endswith('__'):
            return super(RequestAliasedSource, self).__getattribute__(key)
        if key in ('tablename', 'model'):
            return super(RequestAliasedSource, self).__getattribute__(key)
        c = None if self.model is None else getattr(self.model, key, None)
        return RequestTableBelongedColumn(key, self.tablename, c)


class RequestNamedColumn(OrmComparatorMixin, object):
    def __init__(self, column_name):
        self.column_name = column_name


class RequestJoinAbstract(object):
    def __init__(self, source, *args):
        self.source = source
        self.condition = list(args)
        self.join_type = "JOIN"


class RequestLeftJoin(RequestJoinAbstract):
    """ SQL 'LEFT JOIN' (or just 'JOIN') declaration """
    def __init__(self, model, *args):
        super(RequestLeftJoin, self).__init__(model, *args)
        self.join_type = 'LEFT JOIN'


class RequestRightJoin(RequestJoinAbstract):
    """ SQL 'RIGHT JOIN' declaration """
    def __init__(self, model, *args):
        super(RequestRightJoin, self).__init__(model, *args)
        self.join_type = 'RIGHT JOIN'


class RequestInnerJoin(RequestJoinAbstract):
    """ SQL 'INNER JOIN' declaration """
    def __init__(self, model, *args):
        super(RequestInnerJoin, self).__init__(model, *args)
        self.join_type = 'INNER JOIN'


class RequestFullJoin(RequestJoinAbstract):
    """ SQL 'FULL JOIN' declaration """
    def __init__(self, model, *args):
        super(RequestFullJoin, self).__init__(model, *args)
        self.join_type = 'FULL JOIN'


class DbFunctionOperator(object):
    """ A class used when needed to inform the engine about constant string operator inside a function """
    def __init__(self, text, separated=False):
        self.text = text
        self.separated = separated  # place , in front of text


class DbFunctionAbstract(OrmComparatorMixin, AttributeAbstract):
    """ Abstract class used by function declaration classes """

    def __init__(self):
        super(DbFunctionAbstract, self).__init__()
        self.l10n = False

    def __hash__(self):
        return super(OrmComparatorMixin, self).__hash__()

    def resolve(self, context, table_name=None):
        pass

    def as_(self, name):
        """ Returns this instance formatted to be named column as 'name'. """
        from .attributes.base import AttributeAliased
        return AttributeAliased(self, name)

    def text_repr(self, value, instance):
        return self._l10n_text(value)

    def _l10n_text(self, s):
        """ Tries to localize given text. """
        from .utils import gettext
        if isinstance(self.l10n, bool):
            return s if not self.l10n else gettext(s)
        return self.l10n(s) if callable(self.l10n) else s

    @staticmethod
    def _resolve_arg(arg, context, table_name=None):
        from .request import OrmRequest
        from .attributes.base import ColumnAttributeAbstract

        _params = context.params if isinstance(context, OrmRequest) else context
        if isinstance(arg, UncastableColumn):
            arg = arg.c
        elif isinstance(arg, ColumnAttributeAbstract) and context.request_type == 'select':
            arg = arg.make_read(context)
        if isinstance(arg, RawSql):
            return [str(arg.value), True]
        elif isinstance(arg, OrmRequest):
            query_ = arg.sql
            params_ = arg.params
            for k in params_:
                v = params_[k]
                k_ = 'p%i' % len(_params)
                kf_ = "%(" + k + ")s"
                kt_ = "%(" + k_ + ")s"
                query_ = query_.replace(kf_, kt_)
                _params[k_] = v
            return ["(" + query_ + ")", True]
        elif isinstance(arg, (ColumnAttributeAbstract, RequestNamedColumn)) and isinstance(context, OrmRequest):
            context.ensure_column(arg)
            return [context.sql_column(arg, alias=False, table_name=table_name), True]
        elif isinstance(arg, RequestTableBelongedColumn):
            return [context.resolve_expr(arg, allow_alias=False, table_name=table_name), True]
        elif isinstance(arg, DbFunctionAbstract):
            return [arg.resolve(context, table_name=table_name), True]
        elif isinstance(arg, DbFunctionOperator):
            return [arg.text, bool(arg.separated)]
        elif isinstance(arg, OrmComparator) and isinstance(context, OrmRequest):
            return [context.resolve_expr(arg, table_name=table_name), True]
        elif arg is None:
            return ['NULL', True]
        else:
            pn = 'p%i' % len(_params)
            an = "%(" + pn + ")s"
            _params[pn] = arg
            return [an, True]


class DbSimpleFunction(DbFunctionAbstract):
    """ NAME(arg1, arg2, arg*) """
    def __init__(self, funcname, *args, **kwargs):
        super(DbSimpleFunction, self).__init__()
        self.funcname = funcname
        self.args = args or []
        self.pre = list()
        self.post = list()
        self.modifiers = kwargs

    def resolve(self, context, table_name=None):
        """ Returns textual SQL representation of function. """
        args = []
        for arg in self.args:
            args.append(self._resolve_arg(arg, context, table_name=table_name))
        ra = ""
        if args:
            for arg in args:
                if ra and bool(arg[1]):
                    ra += ', '
                elif ra:
                    ra += ' '
                ra += str(arg[0])
        if self.modifiers:
            for arg_name in self.modifiers:
                arg_value = self.modifiers[arg_name]
                if ra:
                    ra += ' '
                ra += arg_name.upper() + ' ' + str(arg_value)
        expr_ = list()
        if self.pre:
            expr_.append(" ".join(self.pre))
        expr_.append(ra)
        if self.post:
            expr_.append(" ".join(self.post))
        return self.funcname + "(" + " ".join(expr_) + ")"


class DbCaseFunction(DbFunctionAbstract):
    """
    CASE value WHEN cases1[0] THEN cases1[1] WHEN cases2[0] THEN cases2[1] ELSE default END
    CASE WHEN cases1[0] THEN cases1[1] WHEN cases2[0] THEN cases2[1] ELSE default END
    """
    def __init__(self, has_else, value, default, *args):
        super(DbCaseFunction, self).__init__()
        self.value = value
        self.cases = args
        self.has_else = bool(has_else)
        self.default = default

    def resolve(self, context, table_name=None):
        """ Returns textual SQL representation of function. """
        csr = list()
        for cs in self.cases:
            if not isinstance(cs, (list, tuple)) or len(cs) != 2:
                raise ValueError(
                    "'case_...' functions requires that 'cases' to be set using list format"
                    " [WHEN expr, THEN expr], [WHEN expr, THEN expr]...!"
                )
            cs_when = self._resolve_arg(cs[0], context, table_name=table_name)
            cs_then = self._resolve_arg(cs[1], context, table_name=table_name)
            csv = "WHEN " + str(cs_when[0]) + " THEN " + str(cs_then[0])
            csr.append(csv)
        csr = " ".join(csr)
        if self.value is not None:
            value = self._resolve_arg(self.value, context, table_name=table_name)
            r = "CASE " + str(value[0]) + " " + csr
        else:
            r = "CASE " + csr
        if self.has_else:
            default = self._resolve_arg(self.default, context, table_name=table_name)
            r += " ELSE " + str(default[0])
        r += " END"
        return "(" + r + ")"


class DbMiddleFunction(DbFunctionAbstract):
    """ left NAME right [ex] """
    def __init__(self, name, left, right, ex=None):
        super(DbMiddleFunction, self).__init__()
        self.name = name
        self.left = left
        self.right = right
        self.ex = ex

    def resolve(self, context, table_name=None):
        """ Returns textual SQL representation of function. """
        larg_ = self._resolve_arg(self.left, context, table_name=table_name)
        rarg_ = self._resolve_arg(self.right, context, table_name=table_name)
        r = str(larg_) + " " + str(self.name) + " " + str(rarg_)
        if self.ex is not None:
            exarg_ = self._resolve_arg(self.ex, context, table_name=table_name)
            r += " " + str(exarg_)
        return r


def select_(*args, **kwargs):
    """ Helper function used to generate a SELECT request using given parameters. A shortcut
    for OrmSelectRequest's class instance creating. """
    from .request import OrmSelectRequest
    return OrmSelectRequest(*args, **kwargs)


def uncastable_(c):
    """ This function returns a wrapped uncastable column that used in functions when declaring
    _attrs which are castable itself, to prevent recursion when function resolved column which
    returns a function which resolves an argument which is castable column again and again
    and again. """
    return UncastableColumn(c)


def raw_(sql):
    """ Used to inform the engine that this is a textual name of the requesting column
    or other textual raw query """
    return RawSql(sql)


def left_join_(cls, *args):
    """ Returns LEFT JOIN instance """
    return RequestLeftJoin(cls, *args)


def join_(cls, *args):
    """ Returns LEFT JOIN instance """
    return left_join_(cls, *args)


def right_join_(cls, *args):
    """ Returns RIGHT JOIN instance """
    return RequestRightJoin(cls, *args)


def inner_join_(cls, *args):
    """ Returns INNER JOIN instance """
    return RequestInnerJoin(cls, *args)


def full_join_(cls, *args):
    """ Returns FULL JOIN instance """
    return RequestFullJoin(cls, *args)


def t_(table, model=None):
    """ Offers a source by name of table alias. """
    from .models.base import OrmModelAliased
    if isinstance(table, OrmModelAliased):
        table_name = table.alias
        model = model if model is not None else table.model
    else:
        table_name = table
    return RequestAliasedSource(table_name, model)


def c_(column_name):
    """ Offers a column by name of column alias. """
    return RequestNamedColumn(column_name)


def and_(*args):
    """ SQL 'AND' operator """
    return ['AND'] + list(args)


def or_(*args):
    """ SQL 'OR' operator """
    return ['OR'] + list(args)


def not_(expr):
    """ SQL 'NOT' operator """
    return func_('NOT', expr)


def filter_(*args):
    """ Is synonym to 'where_' function """
    return RequestWhere(*args)


def where_(*args):
    """ Defines an SQL 'WHERE' condition expression """
    return RequestWhere(*args)


def having_(*args):
    """ Defines an SQL 'HAVING' condition expression """
    return RequestHaving(*args)


def limit_(*args):
    """ Defines SQL 'LIMIT' operator """
    return RequestLimit(*args)


def order_by_(c, d='ASC'):
    """ Defines SQL 'ORDER BY' operator """
    return RequestOrderBy(c, d)


def group_by_(c):
    """ Defines SQL 'GROUP BY' operator """
    return RequestGroupBy(c)


def textual_(c):
    """ Resolves the column into textual representation if 'options' parameter of the
    attribute is set. """
    options = getattr(c, 'options', None)
    if not options:
        return concat_(c, '')
    if not isinstance(options, dict):
        raise OrmDefinitionError("attribute's 'options' must be a (dict) type!")
    exprs = list()
    for k in options:
        v = options[k]
        exprs.append([k, v])
    return case_else_(c, concat_(c, ''), *exprs)


def func_(name, *args, **kwargs):
    """ General (any custom) SQL function declaration """
    if not isinstance(name, str):
        raise ValueError("'func_' require that 'name' must be a string!")
    name = name.upper()
    return DbSimpleFunction(name, *args, **kwargs)


def isnull_(expr):
    """ ISNULL(expr) """
    return func_('ISNULL', expr)


def exists_(expr):
    """ EXISTS(expr) """
    return func_('EXISTS', expr)


def case_(value, *args):
    """ CASE value WHEN a1 THEN b1 WHEN a2 THEN b2 END """
    return DbCaseFunction(False, value, None, *args)


def case_else_(value, default, *args):
    """ CASE value WHEN a1 THEN b1 WHEN a2 THEN b2 ELSE default END """
    return DbCaseFunction(True, value, default, *args)


def case_when_(*args):
    """ CASE WHEN a1 THEN b1 WHEN a2 THEN b2 END """
    return case_(None, *args)


def case_when_else_(default, *args):
    """ CASE WHEN a1 THEN b1 WHEN a2 THEN b2 ELSE default END """
    return case_else_(None, default, *args)


def if_(expr1, expr2, expr3):
    """ IF(expr1, expr2, expr3) """
    return func_('IF', expr1, expr2, expr3)


def ifnull_(expr1, expr2):
    """ IFNULL(expr1, expr2) """
    return func_('IFNULL', expr1, expr2)


def nullif_(expr1, expr2):
    """ NULLIF(expr1, expr2) """
    return func_('NULLIF', expr1, expr2)


def abs_(x):
    """ ABS(x) """
    return func_('ABC', x)


def asin_(x):
    """ ASIN(x) """
    return func_('ASIN', x)


def atan_(x):
    """ ATAN(x) """
    return func_('ATAN', x)


def atan2_(y, x):
    """ ATAN2(y, x) """
    return func_('ATAN2', y, x)


def ceiling_(x):
    """ CEILING(x) """
    return func_('CEILING', x)


def ceil_(x):
    """ Alias for 'ceiling' """
    return ceiling_(x)


def conv_(n, from_base, to_base):
    """ CONV(n, from_base, to_base) """
    return func_('CONV', n, from_base, to_base)


def cos_(x):
    """ COS(x) """
    return func_('COS', x)


def cot_(x):
    """ COT(x) """
    return func_('COT', x)


def crc32_(expr):
    """ CRC32(expr) """
    return func_('CRC32', expr)


def degrees_(x):
    """ DEGREES(x) """
    return func_('DEGREES', x)


def exp_(x):
    """ EXP(x) """
    return func_('EXP', x)


def floor_(x):
    """ FLOOR(x) """
    return func_('FLOOR', x)


def format_(x, d, locale=None):
    """
    FORMAT(x, d)
    FORMAT(x, d, locale)
    """
    return func_('FORMAT', x, d) if locale is None else func_('FORMAT', x, d, locale)


def hex_(n):
    """ HEX(n) """
    return func_('HEX', n)


def ln_(x):
    """ LN(x) """
    return func_('LN', x)


def log_(x, y=None):
    """
    LOG(x)
    LOG(x, y)
    """
    return func_('LOG', x) if y is None else func_('LOG', x, y)


def log2_(x):
    """ LOG2(x) """
    return func_('LOG2', x)


def log10_(x):
    """ LOG10(x) """
    return func_('LOG10', x)


def mod_(n, m):
    """ MOD(n, m) """
    return func_('MOD', n, m)


def pi_():
    """ PI() """
    return func_('PI')


def pow_(x, y):
    """ POW(x, y) """
    return func_('POW', x, y)


def power_(x, y):
    """ Alias for 'pow' """
    return pow_(x, y)


def radians_(x):
    """ RADIANS(x) """
    return func_('RADIANS', x)


def rand_(n=None):
    """
    RAND()
    RAND(n)
    """
    return func_('RAND') if n is None else func_('RAND', n)


def round_(x, d=None):
    """
    ROUND(x)
    ROUND(x, d)
    """
    return func_('ROUND', x) if d is None else func_('ROUND', x, d)


def sign_(x):
    """ SIGN(x) """
    return func_('SIGN', x)


def sin_(x):
    """ SIN(x) """
    return func_('SIN', x)


def sqrt_(x):
    """ SQRT(x) """
    return func_('SQRT', x)


def tan_(x):
    """ TAN(x) """
    return func_('TAN', x)


def truncate_(x, d):
    """ TRUNCATE(x, d) """
    return func_('TRUNCATE', x, d)


def ascii_(s):
    """ ASCII(s) """
    return func_('ASCII', s)


def bin_(n):
    """ BIN(n) """
    return func_('BIN', n)


def bin_length_(s):
    """ BIN_LENGTH(s) """
    return func_('BIN_LENGTH', s)


def char_(*args, **kwargs):
    """
    CHAR(arg1, arg2, ...)
    CHAR(arg1, arg2, ... USING using)
    """
    kw = {}
    if 'using' in kwargs:
        kw['USING'] = kwargs.pop('using')
    return func_('CHAR', *args, **kw)


def char_length_(s):
    """ CHAR_LENGTH(s) """
    return func_('CHAR_LENGTH', s)


def character_length_(s):
    """ Alias for 'char_length' """
    return char_length_(s)


def concat_(*args):
    """ CONCAT(arg1, arg2, ...) """
    return func_('CONCAT', *args)


def concat_ws_(separator, *args):
    """ CONCAT_WS(separator, arg1, arg2, ...) """
    return func_('CONCAT_WS', separator, *args)


def elt_(n, *args):
    """ ELT(n, arg1, arg2, ...) """
    return func_('ELT', n, *args)


def export_set_(bits, on, off, separator=None, number_of_bits=None):
    """
    EXPORT_SET(bits, on, off)
    EXPORT_SET(bits, on, off, separator)
    EXPORT_SET(bits, on, off, separator, number_of_bits)
    """
    if number_of_bits is not None:
        return func_("EXPORT_SET", bits, on, off, separator, number_of_bits)
    elif separator is not None:
        return func_("EXPORT_SET", bits, on, off, separator)
    else:
        return func_("EXPORT_SET", bits, on, off)


def field_(s, *args):
    """ FIELD(s, arg1, arg2, ...) """
    return func_('FIELD', s, *args)


def find_in_set_(s, l):
    """ FIND_IN_SET(s, l) """
    return func_('FIND_IN_SET', s, l)


def from_base64_(s):
    """ FROM_BASE64(s) """
    return func_('FROM_BASE64', s)


def insert_(s, pos, length, newstr):
    """ INSERT(s, pos, length, newstr) """
    return func_('INSERT', s, pos, length, newstr)


def instr_(s, substr):
    """ INSTR(s, substr) """
    return func_("INSTR", s, substr)


def lower_(s):
    """ LOWER(s) """
    return func_('LOWER', s)


def lcase_(s):
    """ Alias for 'lower' """
    return lower_(s)


def left_(s, length):
    """ LEFT(s, length) """
    return func_('LEFT', s, length)


def length_(s):
    """ LENGTH(s) """
    return func_('LENGTH', s)


def locate_(substr, s, pos=None):
    """
    LOCATE(substr, s)
    LOCATE(substr, s, pos)
    """
    return func_('LOCATE', substr, s) if pos is None else func_('LOCATE', substr, s, pos)


def lpad_(s, length, padstr):
    """ LPAD(s, length, padstr) """
    return func_('LPAD', s, length, padstr)


def ltrim_(s):
    """ LTRIM(s) """
    return func_('LTRIM', s)


def make_set(bits, *args):
    """ MAKE_SET(bits, arg1, arg2, ...) """
    return func_('MAKE_SET', bits, *args)


def mid_(s, pos, length):
    """ MID(s, pos, length) """
    return substring_(s, pos, length)


def oct_(n):
    """ OCT(n) """
    return func_('OCT', n)


def octet_length(s):
    """ OCTET_LENGTH(s) """
    return length_(s)


def ord_(s):
    """ ORD(s) """
    return func_('ORD', s)


def position_(substr, s):
    """ Alias for 'locate' """
    return locate_(substr, s)


def quote_(s):
    """ QUOTE(s) """
    return func_('QUOTE', s)


def repeat(s, cnt):
    """ REPEAT(s, cnt) """
    return func_('REPEAT', s, cnt)


def replace_(s, from_str, to_str):
    """ REPLACE(s, from_str, to_str) """
    return func_('REPLACE', s, from_str, to_str)


def reverse_(s):
    """ REVERSE(s) """
    return func_('REVERSE', s)


def right_(s, length):
    """ RIGHT(s, length) """
    return func_('RIGHT', s, length)


def rpad_(s, length, padstr):
    """ RPAD(s, length, padstr) """
    return func_('RPAD', s, length, padstr)


def rtrim_(s):
    """ RTRIM(s) """
    return func_('RTRIM', s)


def soundex(s):
    """ SOUNDEX(s) """
    return func_('SOUNDEX', s)


def space_(n):
    """ SPACE(n) """
    return func_('SPACE', n)


def substr_(s, pos, length=None):
    """ Alias for 'substring' """
    return substring_(s, pos, length)


def substring_(s, pos, length=None):
    """
    SUBSTRING(s, pos)
    SUBSTRING(s, pos, length)
    """
    return func_('SUBSTRING', s, pos) if length is None else func_('SUBSTRING', s, pos, length)


def substring_index_(s, delim, count):
    """ SUBSTRING_INDEX(s, delimn, count) """
    return func_('SUBSTRING_INDEX', s, delim, count)


def to_base64_(s):
    """ TO_BASE64(s) """
    return func_('TO_BASE64', s)


def trim_(s, from_=None):
    """
    TRIM(s)
    TRIM(from_ FROM s)
    """
    return func_('TRIM', s) if from_ is None else func_('TRIM', from_, DbFunctionOperator('FROM'), s)


def trim_leading_(s, from_=None):
    """
    TRIM(LEADING s)
    TRIM(LEADING from_ FROM s)
    """
    if from_ is None:
        return func_('TRIM', DbFunctionOperator('LEADING'), s)
    else:
        return func_('TRIM', DbFunctionOperator('LEADING'), from_, DbFunctionOperator('FROM'), s)


def trim_trailing_(s, from_=None):
    """
    TRIM(TRAILING s)
    TRIM(TRAILING from_ FROM s)
    """
    if from_ is None:
        return func_('TRIM', DbFunctionOperator('TRAILING'), s)
    else:
        return func_('TRIM', DbFunctionOperator('TRAILING'), from_, DbFunctionOperator('FROM'), s)


def trim_both_(s, from_=None):
    """
    TRIM(BOTH s)
    TRIM(BOTH from_ FROM s)
    """
    if from_ is None:
        return func_('TRIM', DbFunctionOperator('BOTH'), s)
    else:
        return func_('TRIM', DbFunctionOperator('BOTH'), from_, DbFunctionOperator('FROM'), s)


def ucase_(s):
    """ Alias for 'upper' """
    return upper_(s)


def unhex_(s):
    """ UNHEX(s) """
    return func_('UNHEX', s)


def upper_(s):
    """ UPPER(s) """
    return func_('UPPER', s)


def strcmp_(s1, s2, collate=None):
    """
    STRCMP(s1, s2)
    STRCMP(s1, s2 COLLATE collate)
    """
    if collate is None:
        return func_('STRCMP', s1, s2)
    else:
        return func_('STRCMP', s1, s2, collate=collate)


def like_(expr, pat):
    """ expr LIKE pat """
    return DbMiddleFunction('LIKE', expr, pat)


def notlike_(expr, pat):
    """ expr NOT LINKE pat """
    return DbMiddleFunction('NOT LIKE', expr, pat)


def rlike_(expr, pat):
    """ expr RLIKE pat """
    return DbMiddleFunction('RLIKE', expr, pat)


def not_rlike_(expr, pat):
    """ expr NOT RLIKE pat """
    return DbMiddleFunction('NOT RLIKE', expr, pat)


def regexp_(expr, pat):
    """ expr REGEXP pat """
    return DbMiddleFunction('REGEXP', expr, pat)


def not_regexp_(expr, pat):
    """ expr NOT REGEXP pat """
    return DbMiddleFunction('NOT REGEXP', expr, pat)


def adddate_(d, interval=None, days=None):
    """
    ADDDATE(d, INTERVAL interval DAYS)
    ADDDATE(d, days)
    """
    if interval is None and days is None:
        raise AttributeError("'ADDDATE' required 'interval' or 'days' to be set!")
    if interval:
        interval_ = "INTERVAL " + interval
        return func_('ADDDATE', d, DbFunctionOperator(interval_, True))
    else:
        return func_('ADDDATE', d, days)


def addtime_(expr1, expr2):
    """ ADDTIME(expr1, expr2) """
    return func_('ADDTIME', expr1, expr2)


def convert_tz_(dt, from_tz, to_tz):
    """ CONVERT_TZ(dt, from_tz, to_tz) """
    return func_('CONVERT_TZ', dt, from_tz, to_tz)


def curdate_():
    """ CURDATE() """
    return func_('CURDATE')


def current_date():
    """ Alias for 'curdate' """
    return curdate_()


def current_time(fsp=None):
    """ Alias for 'curtime' """
    return curtime_(fsp)


def curtime_(fsp=None):
    """
    CURTIME()
    CURTIME(fsp)
    """
    return func_('CURTIME') if fsp is None else func_('CURTIME', fsp)


def date_(expr):
    """ DATE(expr) """
    return func_('DATE', expr)


def datediff_(expr1, expr2):
    """ DATEDIFF(expr1, expr2) """
    return func_('DATEDIFF', expr1, expr2)


def date_add_(d, interval):
    """ DATE_ADD(d INTERVAL interval) """
    return adddate_(d, interval=interval)


def date_sub_(d, interval):
    """ DATE_SUB(d INTERVAL interval) """
    return subdate_(d, interval=interval)


def date_format_(d, fmt):
    """ DATE_FORMAT(d, fmt) """
    return func_('DATE_FORMAT', d, fmt)


def subdate_(d, interval=None, days=None):
    """
    SUBDATE(d, INTERVAL interval)
    SUBDATE(d, days)
    """
    if interval is None and days is None:
        raise AttributeError("'SUBDATE' required 'interval' or 'days' to be set!")
    if interval:
        interval_ = "INTERVAL " + interval
        return func_('SUBDATE', d, DbFunctionOperator(interval_, True))
    else:
        return func_('SUBDATE', d, days)


def dayofmonth_(d):
    """ DAYOFMONTH(d) """
    return func_('DAYOFMONTH', d)


def day_(d):
    """ DAY(d) """
    return dayofmonth_(d)


def dayname_(d):
    """ DAYNAME(d) """
    return func_('DAYNAME', d)


def dayofweek_(d):
    """ DAYOFWEEK(d) """
    return func_('DAYOFWEEK', d)


def dayofyear_(d):
    """ DAYOFYEAR(d) """
    return func_('DAYOFYEAR', d)


def extract_(unit, d):
    """ EXTRACT(unit FROM d) """
    return func_('EXTRACT', unit, DbFunctionOperator('FROM'), d)


def from_days_(n):
    """ FROM_DAYS(n) """
    return func_('FROM_DAYS', n)


def from_unixtime_(unix_timestamp, fmt=None):
    """
    FROM_UNIXTIME(unix_timestamp)
    FROM_UNIXTIME(unix_timestamp, fmt)
    """
    return func_('FROM_UNIXTIME', unix_timestamp) if fmt is None \
        else func_('FROM_UNIXTIME', unix_timestamp, fmt)


def get_format_(type_, target_):
    """ GET_FORMAT(type_, target_) """
    if not isinstance(type_, str):
        raise AttributeError("'GET_FORMAT'('type_') has to be string!")
    if not isinstance(target_, str):
        raise AttributeError("'GET_FORMAT'('target_') has to be a string!")
    type_ = type_.upper()
    target_ = target_.upper()
    if type_ not in ('DATE', 'TIME', 'DATETIME'):
        raise ValueError("'GET_FORMAT'('type_') must be 'DATE', 'TIME' or 'DATETIME'!")
    if target_ not in ('EUR', 'USA', 'JIS', 'ISO', 'INTERNAL'):
        raise ValueError("'GET_FORMAT'('target_') must be 'EUR', 'USA', 'JIS', 'ISO' or 'INTERNAL'!")
    return func_('GET_FORMAT', type_, target_)


def hour_(t):
    """ HOUR(t) """
    return func_('HOUR', t)


def last_day_(d):
    """ LAST_DAY(d) """
    return func_('LAST_DAY', d)


def localtime_(fsp=None):
    """
    LOCALTIME()
    LOCALTIME(fsp)
    """
    return now_(fsp)


def localtimestamp_(fsp=None):
    """
    LOCALTIMESTAMP()
    LOCALTIMESTAMP(fsp)
    """
    return now_(fsp)


def makedate_(year, dayofyear):
    """ MAKEDATE(year, dayofyear) """
    return func_('MAKEDATE', year, dayofyear)


def maketime_(hour, minute, second):
    """ MAKETIME(hour, minute, second) """
    return func_('MAKETIME', hour, minute, second)


def microsecond_(expr):
    """ MICROSECOND(expr) """
    return func_('MICROSECOND', expr)


def minute_(t):
    """ MINUTE(t) """
    return func_('MINTE', t)


def month_(d):
    """ MONTH(d) """
    return func_('MONTH', d)


def monthname_(d):
    """ MONTHNAME(d) """
    return func_('MONTHDATE', d)


def now_(fsp=None):
    """
    NOW()
    NOW(fsp)
    """
    return func_('NOW') if fsp is None else func_('NOW', fsp)


def period_add_(p, n):
    """ PERIOD_ADD(p, n) """
    return func_('PERIOD_ADD', p, n)


def period_diff_(p1, p2):
    """ PERIOD_DIFF(p1, p2) """
    return func_('PERIOD_DEFF', p1, p2)


def quarter_(d):
    """ QUARTER(d) """
    return func_('QUARTER', d)


def second_(t):
    """ SECOND(t) """
    return func_('SECOND', t)


def sec_to_time_(seconds):
    """ SEC_TO_TIME(second) """
    return func_('SEC_TO_TIME', seconds)


def str_to_date_(s, fmt):
    """ STR_TO_DATE(s, fmt) """
    return func_('STR_TO_DATE', s, fmt)


def subtime_(expr1, expr2):
    """ SUBTIME(expr1, expr2) """
    return func_('SUBTIME', expr1, expr2)


def sysdate_(fsp=None):
    """
    SYSDATE()
    SYSDATE(fsp)
    """
    return func_('SYSDATE') if fsp is None else func_('SYSDATE', fsp)


def time_(expr):
    """ TIME(expr) """
    return func_('TIME', expr)


def timediff_(expr1, expr2):
    """ TIMEDIFF(expr1, expr2) """
    return func_('TIMEDIFF', expr1, expr2)


def timestamp_(expr, expr2=None):
    """
    TIMESTAMP(expr)
    TIMESTAMP(expr, expr2)
    """
    return func_('TIMESTAMP', expr) if expr2 is None else func_('TIMESTAMP', expr, expr2)


def timestampadd_(unit, interval, datetime_expr):
    """ TIMESTAMPADD(unit, interval, datetime_expr) """
    return func_('TIMESTAMPADD', DbFunctionOperator(unit), interval, datetime_expr)


def timestampdiff_(unit, datetime_expr1, datetime_expr2):
    """ TIMESTAMPDIFF(unit, datetime_expr1, datetime_expr2) """
    return func_('TIMESTAMPDIFF', DbFunctionOperator(unit), datetime_expr1, datetime_expr2)


def time_format_(t, fmt):
    """ TIME_FORMAT(t, fmt) """
    return func_('TIME_FORMAT', t, fmt)


def time_to_sec_(t):
    """ TIME_TO_SEC(t) """
    return func_('TIME_TO_SEC', t)


def to_days_(d):
    """ TO_DAYS(d) """
    return func_('TO_DAYS', d)


def to_seconds_(expr):
    """ TO_SECONDS(expr) """
    return func_('TO_SECONDS', expr)


def unix_timestamp_(d=None):
    """
    UNIX_TIMESTAMP()
    UNIX_TIMESTAMP(d)
    """
    return func_('UNIX_TIMESTAMP') if d is None else func_('UNIX_TIMESTAMP', d)


def utc_date_():
    """ UTC_DATE() """
    return func_('UTC_DATE')


def utc_time_(fsp=None):
    """
    UTC_TIME()
    UTC_TIME(fsp)
    """
    return func_('UTC_TIME') if fsp is None else func_('UTC_TIME', fsp)


def utc_timestamp_(fsp=None):
    """
    UTC_TIMESTAMP()
    UTC_TIMESTAMP(fsp)
    """
    return func_('UTC_TIMESTAMP') if fsp is None else func_('UTC_TIMESTAMP', fsp)


def week_(d, mode=None):
    """
    WEEK(d)
    WEEK(d, mode)
    """
    return func_('WEEK', d) if mode is None else func_('WEEK', d, mode)


def weekday_(d):
    """ WEEKDAY(d) """
    return func_('WEEKDAY', d)


def weekofyear_(d):
    """ WEEKOFYEAR(d) """
    return func_('WEEKOFYEAR', d)


def year_(d):
    """ YEAR(d) """
    return func_('YEAR', d)


def yearweek_(d, mode=None):
    """
    YEARWEEK(d)
    YEARWEEK(d, mode)
    """
    return func_('YEARWEEK', d) if mode is None else func_('YEARWEEK', d, mode)


def cast_(expr, type_):
    """ CAST(expr AS type_) """
    if not isinstance(type_, str):
        raise ValueError("'CAST'('type_') must be a string!")
    _as = 'AS ' + str(type_)
    return func_('CAST', expr, DbFunctionOperator(_as))


def convert_(expr, type_):
    """ CONVERT(expr, type_) """
    if not isinstance(type_, str):
        raise ValueError("'CONVERT'('type_') must be a string!")
    return func_('CONVERT', expr, type_)


def convert_using_(expr, using):
    """ CONVERT(expr USING using) """
    if not isinstance(using, str):
        raise ValueError("'CONVERT'('using') must be a string!")
    return func_('CONVERT', expr, DbFunctionOperator('USING ' + str(using)))


def binary_(expr):
    """ CAST(expr AS BINARY) """
    return cast_(expr, 'BINARY')


def extractvalue_(xml_frag, xpath_expr):
    """ ExtractValue(xml_frag, xpath_expr) """
    return func_("ExtractValue", xml_frag, xpath_expr)


def updatexml_(xml_target, xpath_expr, new_xml):
    """ UpdateXML(xml_target, xpath_expr, new_xml) """
    return func_("UpdateXML", xml_target, xpath_expr, new_xml)


def bit_count_(n):
    """ BIT_COUNT(n) """
    return func_('BIT_COUNT', n)


def aes_decrypt_(crypt_str, key_str, init_vector=None):
    """
    AES_DECRYPT(crypt_str, key_str)
    AES_DECRYPT(crypt_str, key_str, init_vector)
    """
    return func_('AES_DECRYPT', crypt_str, key_str) if init_vector is None \
        else func_('AES_DECRYPT', crypt_str, key_str, init_vector)


def aes_encrypt_(s, key_str, init_vector=None):
    """
    AES_ENCRYPT(crypt_str, key_str)
    AES_ENCRYPT(crypt_str, key_str, init_vector)
    """
    return func_('AES_ENCRYPT', s, key_str) if init_vector is None \
        else func_('AES_ENCRYPT', s, key_str, init_vector)


def compress_(s):
    """ COMPRESS(s) """
    return func_('COMPRESS', s)


def md5_(s):
    """ MD5(s) """
    return func_('MD5', s)


def random_bytes_(length):
    """ RANDOM_BYTES(length) """
    return func_('RANDOM_BYTES', length)


def sha_(s):
    """ SHA(s) """
    return func_('SHA', s)


def sha1_(s):
    """ Alias for 'sha' """
    return sha_(s)


def sha2_(s, hash_length):
    """ SHA2(s, hash_length) """
    return func_('SHA2', s, hash_length)


def uncompress_(s):
    """ UNCOMPRESS(s) """
    return func_('UNCOMPRESS', s)


def uncompressed_length_(compressed_string):
    """ UNCOMPRESSES_LENGTH(compressed_string) """
    return func_('UNCOMPRESSED_LENGTH', compressed_string)


def validate_password_strength(s):
    """ VALiDATE_PASSWORD_STRENGTH(s) """
    return func_('VALIDATE_PASSWORD_STRENGTH', s)


def any_value_(arg):
    """ ANY_VALUE(arg) """
    return func_('ANY_VALUE', arg)


def default_(col_name):
    """ DEFAULT(col_name) """
    return func_('DEFAULT', col_name)


def inet_aton_(expr):
    """ INET_ATON(expr) """
    return func_('INET_ATON', expr)


def inet_ntoa_(expr):
    """ INET_NTOA(expr) """
    return func_('INET_NTOA', expr)


def inet6_aton_(expr):
    """ INET6_ATON(expr) """
    return func_('INET6_ATON', expr)


def inet6_ntoa_(expr):
    """ INET6_NTOA(expr) """
    return func_('INET6_NTOA', expr)


def is_ipv4_(expr):
    """ IS_IPV4(expr) """
    return func_('IS_IPV4', expr)


def is_ipv4_compat_(expr):
    """ IS_IPV4_COMPAT(expr) """
    return func_('IS_IPV4_COMPAT', expr)


def is_ipv4_mapped_(expr):
    """ IS_IPV4_MAPPED(expr) """
    return func_('IS_IPV4_MAPPED', expr)


def is_ipv6_(expr):
    """ IS_IPV6(expr) """
    return func_('IS_IPV6', expr)


def uuid_():
    """ UUID() """
    return func_('UUID')


def uuid_short_():
    """ UUID_SHORT() """
    return func_('UUID_SHORT')


def avg_(expr, distinct=False):
    """
    AVG(expr)
    AVG(DISTINCT expr)
    """
    return func_('AVG', expr) if not distinct \
        else func_('AVG', DbFunctionOperator('DISTINCT'), expr)


def bit_and_(expr):
    """ BIT_AND(expr) """
    return func_('BIT_AND', expr)


def bit_or_(expr):
    """ BIT_OR(expr) """
    return func_('BIT_OR', expr)


def bit_xor_(expr):
    """ BIT_XOR(expr) """
    return func_('BIT_XOR', expr)


def count_(expr):
    """ COUNT(expr) """
    return func_('COUNT', expr)


def count_distinct_(*args):
    """ COUNT(DISTINCT arg1, arg2, ...) """
    return func_('COUNT', DbFunctionOperator('DISTINCT'), *args)


def max_(expr, distinct=False):
    """
    MAX(expr)
    MAX(DISTINCT expr)
    """
    return func_('MAX', expr) if not distinct \
        else func_('MAX', DbFunctionOperator('DISTINCT'), expr)


def min_(expr, distinct=False):
    """
    MIN(expr)
    MIN(DISTINCT expr)
    """
    return func_('MIN', expr) if not distinct \
        else func_('MIN', DbFunctionOperator('DISTINCT'), expr)


def std_(expr):
    """ STD(expr) """
    return func_('STD', expr)


def stddev_(expr):
    """ STDDEV(expr) """
    return func_('STDDEV', expr)


def stddev_pop_(expr):
    """ STDDEV_POP(expr) """
    return func_('STDDEV_POP', expr)


def stddev_samp_(expr):
    """ STDDEV_SAMP(expr) """
    return func_('STDDEV_SAMP', expr)


def sum_(expr, distinct=False):
    """
    SUM(expr)
    SUM(DISTINCT expr)
    """
    return func_('SUM', expr) if not distinct \
        else func_('SUM', DbFunctionOperator('DISTINCT'), expr)


def var_pop_(expr):
    """ VAR_POP(expr) """
    return func_('VAR_POP', expr)


def var_samp_(expr):
    """ VAR_SAMP(expr) """
    return func_('VAR_SAMP', expr)


def variance_(expr):
    """ VARIANCE(expr) """
    return func_('VARIANCE', expr)


def host_(expr):
    """ PostgreSQL only: HOST(expr) """
    return func_('HOST', expr)


def pg_text_(expr):
    """ PostgresSQL only: TEXT(expr) """
    return func_('TEXT', expr)




from os import makedirs
from os.path import lexists
from inspect import isclass
from .exceptions import OrmManagementError, OrmAttributeTypeMistake
from .storages import OrmFileStorage, OrmPictureStorage


class OrmManager(object):
    def __init__(self):
        self._debug = False
        self._sessions = {}
        self._userident = None
        self._storages = {
            None: {}
        }
        self._journals = {
            None: {}
        }
        self._query_handler = None
        self._model_request_handler = None
        self._model_write_handler = None
        self._gettext = None
        self._storage_base_path = None
        self._logs = {}
        self._telephony = {
            'formats': {
                6: '??-??-??',
                7: '???-??-??',
                10: '(???) ???-??-??',
                11: '+? (???) ???-??-??',
                12: '+?? (???) ???-??-??',
                13: '+??? (???) ???-??-??',
                14: '+?-??? (???) ???-??-??',
            },
            'format_ext': ' (+?)',
            'local_prefixes': {},
        }
        self._finances = {
            'currency_formats': {
                'rur': ('%irub. %dkop.', '%i rub.'),
                'usd': ('$%i.%d', '$%i'),
                'eur': ('%i.%d EURO', '%i EURO')
            },
            'default_currency': False,
        }
        self._datetime_format = {
            'full_date': '%d.%m.%Y',
            'short_date': '%j.%m.%y',
            'full_time': '%H:%i:%s',
            'short_time': '%H:%i'
        }

    def set_debug(self, *args):
        if not args:
            self._debug = False
            return
        if True in args:
            self._debug = True
            return
        elif False in args:
            self._debug = False
            return
        d_ = list()
        for arg in args:
            if not isinstance(arg, str) \
                    or arg.lower() not in ('select', 'modify', 'alter', 'migrate-getstruct', 'query-results'):
                raise OrmManagementError(
                    "set_debug() requires value to be True, False, or a set of actions to debug to: 'select',"
                    " 'modify', 'alter', 'migrate-getstruct'!"
                )
            d_.append(arg.lower())
        self._debug = ",".join(d_)

    def get_debug(self):
        return bool(self._debug) if isinstance(self._debug, bool) else str(str(self._debug) + "!")

    def _get_managing_session(self, server_type, host, admin_username, admin_password, port=None, debug=None):
        if server_type == 'mysql':
            root_database = 'mysql'
        elif server_type == 'postgresql':
            root_database = 'postgres'
        else:
            raise OrmManagementError("server_type must be 'mysql' or 'postgresql'!")
        kw = {
            'server': server_type,
            'dbname': root_database,
            'username': admin_username,
            'password': admin_password,
            'host': host,
            'port': port,
            'debug': debug
        }
        return self.connect_db(**kw)

    @staticmethod
    def connect_mysql(*args, **kwargs):
        from .session import MysqlSession
        return MysqlSession(*args, **kwargs)

    @staticmethod
    def connect_postgresql(*args, **kwargs):
        from .session import PostgresqlSession
        return PostgresqlSession(*args, **kwargs)

    def connect_db(self, *args, **kwargs):
        session_type = kwargs.pop('server', None)
        if not isinstance(session_type, str) or str(session_type).lower() not in ('mysql', 'postgresql'):
            raise OrmManagementError("'server' parameter must be set and be one of 'mysql' or 'postgresql'!")
        if session_type.lower() == 'mysql':
            return self.connect_mysql(*args, **kwargs)
        elif session_type.lower() == 'postgresql':
            return self.connect_postgresql(*args, **kwargs)
        return None

    def is_database_exists(self, database_name, server_type, host, admin_username, admin_password, port=None,
                           session=None, debug=None):
        session_ = session \
                   or self._get_managing_session(server_type, host, admin_username, admin_password, port, debug)
        is_migrating = session_.is_migrating
        session_.is_migrating = True
        r = False
        if server_type == 'mysql':
            dbr = session_.query(
                "SELECT COUNT(1) FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = %(dbname)s",
                {'dbname': database_name}).one()
            r = bool(dbr[0])
        elif server_type == 'postgresql':
            dbr = session_.query(
                "SELECT COUNT(1) FROM pg_database WHERE datname = %(dbname)s",
                {'dbname': database_name}).one()
            r = bool(dbr[0])
        if session is None:
            session_.close()
        session_.is_migrating = is_migrating
        return r

    def drop_database(self, database_name, server_type, host, admin_username, admin_password, port=None, session=None,
                      just_ensure=False, debug=None):
        session_ = session \
                   or self._get_managing_session(server_type, host, admin_username, admin_password, port, debug)
        if session_.dbms_type == 'postgresql':
            from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
            session_.cn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        if self.is_database_exists(database_name=database_name, server_type=server_type, host=host,
                                   admin_username=admin_username, admin_password=admin_password, port=port,
                                   session=session_):
            session_.execute("DROP DATABASE %s" % session_.quote_name(database_name))
        elif not just_ensure:
            raise OrmManagementError("cannot drop database '%s' - is not exists!" % database_name)
        session_.commit()
        if session is None:
            session_.close()

    def create_database(self, database_name, server_type, host, admin_username, admin_password, port=None,
                        session=None, just_ensure=False, debug=None):
        session_ = session \
                   or self._get_managing_session(server_type, host, admin_username, admin_password, port, debug)
        if session_.dbms_type == 'postgresql':
            from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
            session_.cn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        if not self.is_database_exists(database_name=database_name, server_type=server_type, host=host,
                                       admin_username=admin_username, admin_password=admin_password, port=port,
                                       session=session_):
            if server_type == 'mysql':
                session_.execute("CREATE DATABASE `%s` CHARACTER SET utf8 COLLATE utf8_bin" % database_name)
            else:
                session_.execute("CREATE DATABASE %s" % session_.quote_name(database_name))
        elif not just_ensure:
            raise OrmManagementError("cannot create database '%s' - is already exists!" % database_name)
        session_.commit()
        if session is None:
            session_.close()

    def recreate_database(self, database_name, server_type, host, admin_username, admin_password, port=None,
                          session=None, debug=None):
        session_ = session \
                   or self._get_managing_session(server_type, host, admin_username, admin_password, port, debug)
        self.drop_database(database_name=database_name, server_type=server_type, host=host,
                           admin_username=admin_username, admin_password=admin_password, port=port, session=session_,
                           just_ensure=True, debug=debug)
        self.create_database(database_name=database_name, server_type=server_type, host=host,
                             admin_username=admin_username, admin_password=admin_password, port=port, session=session_,
                             just_ensure=True, debug=debug)
        if session is None:
            session_.close()

    def ensure_database(self, database_name, server_type, host, admin_username, admin_password, port=None,
                        session=None, debug=None):
        session_ = session \
                   or self._get_managing_session(server_type, host, admin_username, admin_password, port, debug)
        self.create_database(database_name=database_name, server_type=server_type, host=host,
                             admin_username=admin_username, admin_password=admin_password, port=port, session=session_,
                             just_ensure=True, debug=debug)
        if session is None:
            session_.close()

    def grant_permissions(self, database_name, server_type, host, admin_username, admin_password,
                          user_name, user_password, user_host=None, grant=None, schema='public', port=None,
                          session=None, debug=None):
        session_ = session \
                   or self._get_managing_session(server_type, host, admin_username, admin_password, port, debug)

        if session_.dbms_type == 'postgresql':
            from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
            session_.cn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        if not self.is_database_exists(database_name=database_name, server_type=server_type, host=host,
                                       admin_username=admin_username, admin_password=admin_password, port=port,
                                       session=session_):
            raise OrmManagementError(
                "Cannot grant permissions to database '%s' because it is not exists" % database_name
            )
        if not user_name or not user_password:
            raise OrmManagementError("grant_permissions() requires 'user_name' and 'user_password' to be set!")
        if server_type == 'mysql':
            if not grant:
                grant = 'ALL'
            if not user_host:
                user_host = "localhost"
            session_.execute("GRANT " + grant + " ON `" + database_name +
                             "`.* TO %(user)s@%(host)s IDENTIFIED BY %(passwd)s",
                             {
                                 'grant': grant,
                                 'database': database_name,
                                 'user': user_name,
                                 'passwd': user_password,
                                 'host': user_host
                             })
            session_.commit()
        else:
            pg_role = session_.query("SELECT 1 FROM pg_roles WHERE rolname = %(rolname)s", {'rolname': user_name})
            if not pg_role:
                session_.execute("CREATE USER \"" + user_name + "\" WITH PASSWORD %(passwd)s",
                                 {'passwd': user_password})
            else:
                session_.execute("ALTER ROLE \"" + user_name + "\" WITH PASSWORD %(passwd)s",
                                 {'passwd': user_password})
            if not grant:
                grant = 'ALL'
            session_.execute("GRANT CONNECT, TEMPORARY ON DATABASE \"%s\" TO \"%s\"" % (database_name, user_name),
                             {'database': database_name, 'rolname': user_name})
            session_.commit()
            pg_sess = self.connect_postgresql(host, database_name, admin_username, admin_password, port, debug=debug)
            pg_sess.execute("GRANT %s ON ALL TABLES IN SCHEMA \"%s\" TO \"%s\"" % (grant, schema, user_name))
            pg_sess.execute("GRANT %s ON ALL SEQUENCES IN SCHEMA \"%s\" TO \"%s\"" % (grant, schema, user_name))
            pg_sess.commit()

        if session is None:
            session_.close()

    def set_session_getter(self, f, tag=None):
        if not f or not callable(f):
            raise OrmManagementError("'set_session_getter' expects a function as parameter!")
        self._sessions[tag] = f

    def set_session(self, s, tag=None):
        from .session import OrmSessionAbstract
        if not isinstance(s, OrmSessionAbstract):
            raise OrmManagementError("'set_session' expects a DatabaseSession instance as parameter!")
        self._sessions[tag] = s

    def get_session(self, **kwargs):
        from .session import OrmSessionAbstract
        from .models.base import OrmModelAbstract

        tag_ = None

        session_ = kwargs.pop('session', None)
        if session_ is not None:
            if isinstance(session_, OrmSessionAbstract):
                return session_
            elif callable(session_):
                return session_()

        instance_ = kwargs.pop('instance', None)
        if instance_ is not None:
            if not isinstance(instance_, OrmModelAbstract):
                raise OrmAttributeTypeMistake(
                    "get_session(instance) parameter must be the type of OrmModel or not be set!"
                )
            if instance_.__session__ is not None:
                if isinstance(instance_.__session__, OrmSessionAbstract):
                    return instance_.__session__
                elif callable(instance_.__session__):
                    return instance_.__session__()

        model_ = kwargs.pop('model', None)
        if model_ is None and instance_ is not None:
            model_ = instance_.__class__
        if model_ is not None:
            if not issubclass(model_, OrmModelAbstract):
                raise OrmAttributeTypeMistake(
                    "get_session(model) parameter must be a class based on OrmModel class!"
                )
            if model_.__meta__.session is not None:
                if isinstance(model_.__meta__.session, OrmSessionAbstract):
                    return model_.__meta__.session
                elif callable(model_.__meta__.session):
                    return model_.__meta__.session()
            if model_.__meta__.tag is not None:
                tag_ = model_.__meta__.tag

        if tag_ not in self._sessions:
            return None
        if not self._sessions[tag_]:
            return None
        if isinstance(self._sessions[tag_], OrmSessionAbstract):
            return self._sessions[tag_]
        elif callable(self._sessions[tag_]):
            return self._sessions[tag_]()
        else:
            return None

    def set_user_ident(self, mixed):
        self._userident = mixed

    def get_user_ident(self):
        from .models.meta import OrmModelMeta
        from .models.base import OrmModel
        from .mapper import mapper
        ident = self._userident
        if ident is None and not mapper.is_userident_model_set():
            raise OrmManagementError("user ident must be set before using it in Models!")
        if ident is None:
            ident = mapper.get_userident_model()
        if callable(ident) and not isclass(ident):
            return ident()
        if issubclass(ident, OrmModelMeta) or issubclass(ident, OrmModel):
            return ident
        if isinstance(ident, str):
            model = mapper.get_model_by_name(ident)
            if model is None:
                raise OrmManagementError("user_ident model '%s' must be declared at a time of real queries!" % ident)
            return model
        raise OrmManagementError("user_ident might be set using callable function, Model or model name only!")

    def set_query_handler(self, f):
        if f is None:
            self._query_handler = None
            return
        if not callable(f):
            raise OrmManagementError("'set_query_handler' requires a callable function to be given!")
        self._query_handler = f

    def get_query_handler(self):
        return self._query_handler

    def clear_query_handler(self):
        self.set_query_handler(None)

    def handle_query(self, query, params):
        if not self._query_handler:
            return True
        return self._query_handler(query, params)

    def set_model_request_handler(self, f):
        if f is None:
            self._model_request_handler = None
            return
        if not callable(f):
            raise OrmManagementError("'set_model_request_handler' requires a callable function to be given!")
        self._model_request_handler = f

    def get_model_request_handler(self):
        return self._model_request_handler

    def clear_model_request_handler(self):
        self.set_model_request_handler(None)

    def handle_model_request(self, request):
        if not self._model_request_handler:
            return True
        return self._model_request_handler(request)

    def set_model_write_handler(self, f):
        if f is None:
            self._model_write_handler = None
            return
        if not callable(f):
            raise OrmManagementError("'set_model_write_handler' requires a callable function to be given!")
        self._model_write_handler = f

    def get_model_write_handler(self):
        return self._model_write_handler

    def clear_model_write_handler(self):
        self.set_model_write_handler(None)

    def handle_model_write(self, instance, is_extender, insert_ignore):
        if not self._model_write_handler:
            return True
        return self._model_write_handler(instance, is_extender=is_extender, insert_ignore=insert_ignore)

    def set_gettext(self, f):
        self._gettext = f

    def get_gettext_callback(self):
        return self._gettext

    def set_storage_base_path(self, path):
        self._storage_base_path = path
        if not lexists(path):
            makedirs(path)

    def get_storage_base_path(self):
        return self._storage_base_path

    def set_telephony_formats(self, formats):
        if formats is not None and not isinstance(formats, (list, tuple)):
            raise OrmManagementError("'set_telephony_formats' value must be a type of list or tuple!")
        formats_ = dict()
        for f in formats:
            fl = len(f) - len(f.replace('?', ''))
            formats_[fl] = f
        self._telephony['formats'] = formats_

    def update_telephony_formats(self, formats):
        if formats is not None and not isinstance(formats, (list, tuple)):
            raise OrmManagementError("'update_telephony_formats' value must be a type of list or tuple!")
        formats_ = dict()
        for f in formats:
            fl = len(f) - len(f.replace('?', ''))
            formats_[fl] = f
        for k in formats_:
            self._telephony['formats'][k] = formats_[k]

    def get_telephony_formats(self):
        return self._telephony['formats']

    def set_telephony_format_ext(self, format_ext):
        self._telephony['format_ext'] = format_ext

    def get_telephony_format_ext(self):
        return self._telephony['format_ext']

    def set_telephony_local_prefixes(self, prefixes):
        if not isinstance(prefixes, dict):
            raise OrmManagementError("'set_telephony_local_prefixes' requires value to be a type of (dict) in"
                                     " ({<base_length>: <prefix>, <base_length>: <prefix>})")
        prefixes_ = dict()
        for k in prefixes:
            if not isinstance(k, int):
                raise OrmManagementError("'set_telephony_local_prefixes' requires value to be a type of (dict) in"
                                         " ({<base_length>: <prefix>, <base_length>: <prefix>})")
            prefixes_[k] = str(prefixes[k])
        self._telephony['local_prefixes'] = prefixes_

    def update_telephony_local_prefixes(self, prefixes):
        if not isinstance(prefixes, dict):
            raise OrmManagementError("'set_telephony_local_prefixes' requires value to be a type of (dict) in"
                                     " ({<base_length>: <prefix>, <base_length>: <prefix>})")
        for k in prefixes:
            if not isinstance(k, int):
                raise OrmManagementError("'set_telephony_local_prefixes' requires value to be a type of (dict) in"
                                         " ({<base_length>: <prefix>, <base_length>: <prefix>})")
            self._telephony['local_prefixes'][k] = str(prefixes[k])

    def get_telephony_local_prefixes(self):
        return self._telephony['local_prefixes']

    def set_datetime_format(self, full_date=None, short_date=None, full_time=None, short_time=None):
        if full_date:
            self._datetime_format['full_date'] = str(full_date)
        if short_date:
            self._datetime_format['short_date'] = str(short_date)
        if full_time:
            self._datetime_format['full_time'] = str(full_time)
        if short_time:
            self._datetime_format['short_time'] = str(short_time)

    def get_datetime_formats(self):
        return self._datetime_format

    def get_full_date_format(self, session=None):
        return self._datetime_format['full_date'] if (session is None or session.dt_format is None) \
            else (session.dt_format['full_date'] or self._datetime_format['full_date'])

    def get_short_date_format(self, session=None):
        return self._datetime_format['short_date'] if (session is None or session.dt_format is None) \
            else (session.dt_format['short_date'] or self._datetime_format['short_date'])

    def get_full_time_format(self, session=None):
        return self._datetime_format['full_time'] if (session is None or session.dt_format is None) \
            else (session.dt_format['full_time'] or self._datetime_format['full_time'])

    def get_short_time_format(self, session=None):
        return self._datetime_format['short_time'] if (session is None or session.dt_format is None) \
            else (session.dt_format['short_time'] or self._datetime_format['short_time'])

    def set_default_currency(self, value):
        if not isinstance(value, str) and value is not False:
            raise OrmManagementError("'set_default_currency' requires (str) name of the currency!")
        self._finances['default_currency'] = value

    def set_currency_format(self, currency, full_format, integer_format):
        if not isinstance(currency, str) or not isinstance(full_format, str) or not isinstance(integer_format, str):
            raise OrmManagementError("currency name and formats must be a type of string!")
        self._finances['currency_formats'][currency] = (full_format, integer_format)

    def get_currency_format(self, currenty):
        if currenty not in self._finances['currency_formats']:
            return None
        return self._finances['currency_formats'][currenty]

    def get_default_currency(self):
        return self._finances['default_currency']

    def set_file_storage(self, name, base_path, tag=None, **kwargs):
        if tag not in self._storages:
            self._storages[tag] = dict()
        kwargs['base_path'] = base_path
        self._storages[tag][name] = OrmFileStorage(**kwargs)
        return self._storages[tag][name]

    def set_picture_storage(self, name, base_path, tag=None, **kwargs):
        if tag not in self._storages:
            self._storages[tag] = dict()
        kwargs['base_path'] = base_path
        self._storages[tag][name] = OrmPictureStorage(**kwargs)
        return self._storages[tag][name]

    def set_default_file_storage(self, base_path, tag=None, **kwargs):
        kwargs['base_path'] = base_path
        return self.set_file_storage('__defaultfilestorage__', tag=tag, **kwargs)

    def set_default_picture_storage(self, base_path, tag=None, **kwargs):
        kwargs['base_path'] = base_path
        return self.set_picture_storage('__defaultpicturestorage__', tag=tag, **kwargs)

    def set_default_storage(self, base_path, tag=None, **kwargs):
        kwargs['base_path'] = base_path
        kwargs['tag'] = tag
        self.set_default_file_storage(**kwargs)
        self.set_default_picture_storage(**kwargs)

    def get_storage(self, name, tag=None):
        if tag not in self._storages:
            raise OrmManagementError("tagged storage tree is not exists for tag: '%r'!" % tag)
        if name not in self._storages[tag]:
            raise OrmManagementError("storage is not exists: '%r'!" % name)
        return self._storages[tag][name]

    def get_default_file_storage(self, tag=None):
        return self.get_storage('__defaultfilestorage__', tag=tag)

    def get_default_picture_storage(self, tag=None):
        return self.get_storage('__defaultpicturestorage__', tag=tag)

    @staticmethod
    def _create_journal_model(name, tag=None):
        from .models.base import OrmModel
        from .models.templates import use_model_template
        from .logs import OrmJournalTemplate
        attrs = dict()
        if tag is not None:
            attrs['__tag__'] = tag
        return use_model_template(type(name, (OrmModel,), attrs), OrmJournalTemplate)

    def set_journal(self, name, tag=None, model=None, tablename=None):
        from .mapper import mapper
        from .models.base import OrmModel
        if tag in self._journals and name in self._journals[tag]:
            raise OrmManagementError("journal already exists with this tag: '%s'!" % name)
        if isinstance(model, str):
            model_ = mapper.get_model_by_name(model)
            if model_ is None:
                model_ = self._create_journal_model(model_)
        elif model is None:
            if name != '__default__':
                name_ = ("JournalLog_%s" % name) if tag is None else ("JournalLog_%s_%s" % (name, tag))
            else:
                name_ = "JournalLog" if tag is None else "JournalLog__%s" % tag
            model_ = self._create_journal_model(name_)
        elif issubclass(model, OrmModel):
            if not mapper.is_model_templated_with(model.__name__, 'OrmJournalTemplate'):
                raise OrmManagementError(
                    "Model which about to be used as journal must be declared using template OrmJournalTemplate!"
                )
            model_ = model
        else:
            raise OrmManagementError(
                "'set_journal' requires that 'model' parameter to be None or a type of Model or (str) - model name!"
            )
        model_.__meta__.tablename = tablename
        if tag not in self._journals:
            self._journals[tag] = dict()
        self._journals[tag][name] = model_

    def get_journal(self, name, tag=None):
        if tag not in self._journals or name not in self._journals[tag]:
            return None
        return self._journals[tag][name]

    def unset_journal(self, name, tag=None):
        from .mapper import mapper
        if not self.has_journal(name, tag):
            raise OrmManagementError("journal '%s' not been defined!" % name)
        model = self._journals[tag][name]
        mapper.unmap_model(model)

    def has_journal(self, name, tag=None):
        return tag in self._journals and name in self._journals[tag]

    def set_default_journal(self, tag=None, model=None, tablename=None):
        if self.has_journal('__default__', tag):
            self.unset_journal('__default__', tag)
        self.set_journal('__default__', tag, model, tablename)

    def get_default_journal(self, tag=None):
        return self.get_journal('__default__', tag)

    def unset_default_journal(self, tag=None):
        if not self.has_journal('__default__', tag):
            return
        self.unset_journal('__default__', tag)


manager = OrmManager()
